"""
Here we create a custom plugin that lets users trigger an extra shutdown call to
the executor, giving them control over the ``wait`` and ``cancel_futures``
values that are passed to the method.

This plugin demonstrates how to setup and use a ``pre_exit`` function in a
plugin's entry point object.

These tests have some mild race conditions in them because we're trying to
inspect the intermediate states of tasks. If this becomes too problematic, we
could eliminate at least a few of them by forcing more controlled
synchronization between the tasks and the test process. This would be easiest to
do with a threaded executor instead of a multiprocessing one. The known race
conditions can cause false test failures, but they shouldn't cause incorrect
test successes.
"""

import platform
import time
from concurrent.futures import ProcessPoolExecutor
from functools import cached_property

import pytest
from pyrseus.ctx.api import ExecutorPluginEntryPoint, extract_keywords
from pyrseus.ctx.registry import _CONCURRENT_ENTRY_POINTS, register_plugin
from pyrseus.ctx.mgr import ExecutorCtx


def sleep01_and_retV():
    """
    This is the task function we'll be running. It takes a small but non-trivial
    time to run.
    """
    time.sleep(0.1)
    return "V"


EPS = 0.01


def get_abbrev_state(fut):
    """
    Summarizes the state of a Future.
    """
    run = fut.running()
    cancelled = fut.cancelled()
    done = fut.done()
    val = None if not done or cancelled else fut.result(0)
    return {
        # We only enumerate the combinations we actually expect to see in these
        # tests.
        (False, False, False, None): "pending",
        (True, False, False, None): "running",
        (False, True, False, None): "cancel_pending",
        (False, True, True, None): "cancelled",
        (False, False, True, "V"): "success",
    }[run, cancelled, done, val]


def summarize_states(futs):
    return {get_abbrev_state(fut) for fut in futs}


def wait_for_futs(futs, timeout_secs):
    t0 = time.time()
    while time.time() - t0 < timeout_secs:
        summaries = summarize_states(futs)
        if len({"pending", "running"} & summaries) == 0:
            break
        time.sleep(0.1)


@pytest.fixture
def entry_point_and_futs_list():
    futs = []

    class OurEntryPoint(ExecutorPluginEntryPoint):
        supports_serial = False
        supports_concurrent = True
        is_available = True

        @cached_property
        def allowed_keywords(self):
            return extract_keywords(type(self).create) | extract_keywords(
                ProcessPoolExecutor
            )

        def create(
            self,
            max_workers,
            # These two are the features we're adding to this ad-hoc plugin.
            cancel_futures=False,
            wait=False,
            # These are just to make the testing easier.
            states_before_shutdown=None,
            states_after_shutdown=None,
            # Pass through everything else.
            **kwargs,
        ):
            # Do *not* pass the four new arguments to ProcessPoolExecutor. We
            # just need them to be captured by pre_exit in a closure (happens
            # automatically in Python).
            exe = ProcessPoolExecutor(max_workers, **kwargs)

            # This function will be called just before
            # ProcessPoolExecutor.__exit__. The four extra create() arguments
            # are captured in a closure for it to use.
            def pre_exit(exe, exc_info):
                # The asserts are just for this unit test. If we were creating a
                # real plugin, we'd omit them.
                assert summarize_states(futs) == states_before_shutdown
                # This is the extra feature that this ad-hoc plugin gives us.
                if cancel_futures or wait:
                    exe.shutdown(wait=wait, cancel_futures=cancel_futures)
                # Another assert, just for testing.
                assert summarize_states(futs) == states_after_shutdown

            return exe, pre_exit

    assert "altshutdown" not in _CONCURRENT_ENTRY_POINTS
    register_plugin("altshutdown", OurEntryPoint())
    assert "altshutdown" in _CONCURRENT_ENTRY_POINTS
    yield OurEntryPoint(), futs
    del _CONCURRENT_ENTRY_POINTS["altshutdown"]


def test_dont_use_new_feature(entry_point_and_futs_list):
    _, futs = entry_point_and_futs_list
    # Everything works as usual if we disable the new features.
    with ExecutorCtx(
        "altshutdown",
        1,
        cancel_futures=False,
        wait=False,
        states_before_shutdown={"running", "pending"},
        states_after_shutdown={"running", "pending"},  # custom shutdown = noop
    ) as exe:
        t0 = time.time()
        for _ in range(10):
            futs.append(exe.submit(sleep01_and_retV))
    elapsed_secs = time.time() - t0
    assert elapsed_secs >= (0.1 * 10) - EPS  # all of them ran...
    assert summarize_states(futs) == {"success"}  # ...to completion


def test_wait_only_does_nothing_new(entry_point_and_futs_list):
    _, futs = entry_point_and_futs_list
    # Pre-waiting doesn't accomplish anything new; it just makes the waiting
    # happen in our shutdown callback instead of in
    # ProcessPoolExecutor.__exit__.
    with ExecutorCtx(
        "altshutdown",
        1,
        cancel_futures=False,  # the default
        wait=True,  # what __exit__ does
        states_before_shutdown={"running", "pending"},
        states_after_shutdown={"success"},  # resolved earlier
    ) as exe:
        t0 = time.time()
        for _ in range(10):
            futs.append(exe.submit(sleep01_and_retV))
    elapsed_secs = time.time() - t0
    assert elapsed_secs >= (0.1 * 10) - EPS  # all of them ran...
    assert summarize_states(futs) == {"success"}  # ...and are still completed


def test_cancel_without_wait_does_funny_stuff(entry_point_and_futs_list):
    _, futs = entry_point_and_futs_list
    # Cancelling the futures but not waiting does funny things.
    #  - It disables the waiting not just for the explicit shutdown call, but
    #    also permanently disables it for all subsequent calls, including in
    #    ProcessPoolExecutor.__exit__.
    #  - No futures actually get cancelled.
    #  - The worker pool remains alive, even after ProcessPoolExecutor.__exit__
    #    is called. All of the tasks eventually get completed, unprotected by
    #    the context manager.
    #
    # If any of the above-described behavior changes in a future version of
    # Python, we'll need to adjust this test.
    with ExecutorCtx(
        "altshutdown",
        1,
        cancel_futures=True,  # start cancelling
        wait=False,  # this effectively disables the wait in __exit__
        states_before_shutdown={"running", "pending"},  # running + pending
        states_after_shutdown={"running", "pending"},  # cancels not registered yet
    ) as exe:
        t0 = time.time()
        for _ in range(10):
            futs.append(exe.submit(sleep01_and_retV))
    elapsed_secs = time.time() - t0
    assert elapsed_secs < 0.1  # exited quickly
    assert summarize_states(futs) == {"running", "pending"}  # exited too fast
    # everything keeps working till all tasks are done, despite exiting the
    # context
    wait_for_futs(futs, 1.5 if platform.system() == "Linux" else 5)
    assert summarize_states(futs) == {"success"}


def test_cancel_and_wait_works(entry_point_and_futs_list):
    _, futs = entry_point_and_futs_list
    # Cancelling and waiting in the same shutdown call does what we expect: all
    # running tasks are completed, and all pending ones are  permanently
    # cancelled.
    with ExecutorCtx(
        "altshutdown",
        1,
        cancel_futures=True,  # start cancelling
        wait=True,  # but wait until all running ones are done
        states_before_shutdown={"running", "pending"},  # running + pending
        states_after_shutdown={"success", "cancelled"},  # cancelling worked
    ) as exe:
        t0 = time.time()
        for _ in range(10):
            futs.append(exe.submit(sleep01_and_retV))
    t1 = time.time()
    assert 0.01 - EPS <= t1 - t0  # at least 1 should have finished
    # before all would have finished, we should have a combination of success
    # and cancel results
    wait_for_futs(futs, 0.35 if platform.system() == "Linux" else 0.9)
    assert summarize_states(futs) == {"success", "cancelled"}



"""
Defines the plugin API used by `~pyrseus.ctx.mgr.ExecutorCtx` to register
executor backends.

Each module in the `pyrseus.ctx.plugins` packages must define an ``ENTRY_POINT``
global variable that implements the `.ExecutorPluginEntryPoint` protocol. Third
party plugins must also supply an entry point object, but they are freer to
choose how to have them registered.
"""
from __future__ import annotations

from concurrent.futures import Executor
from enum import auto
from functools import cached_property
from inspect import Parameter, signature
from types import TracebackType
from typing import Callable, List, Optional, Protocol, Set, Tuple, Type, Union

try:
    from enum import StrEnum as _StrishEnum  # >= py 3.11
except ImportError:
    from enum import Enum as _StrishEnum  # < py 3.11

ExcInfo = Tuple[Optional[Type], Optional[BaseException], Optional[TracebackType]]
"""
Formal type for the return value of ``sys.exc_info()``.
"""

PreExitFunc = Callable[[Executor, ExcInfo], None]
"""
Formal type for a "pre exit" function that a plugin's ``ENTRY_POINT.create``
method can return. See `.ExecutorPluginEntryPoint.create` for details.
"""


class OnError(_StrishEnum):
    """
    Enum used by `.ExecutorPluginEntryPoint.create` to indicate how the user
    would like workers and pending futures to be handled when an uncaught
    exception is passing through the context's ``__exit__`` method.

    >>> OnError.get("kill")
    <OnError.KILL_WORKERS...>

    >>> OnError.get(OnError.KILL)
    <OnError.KILL_WORKERS...>
    """

    WAIT = auto()
    """
    Wait on all pending futures to enter a final state. This is the default
    behavior for most existing executors.
    """

    CANCEL_FUTURES = auto()
    """
    Cancel as many pending futures as possible. Note that for some executors
    like |loky|_'s pre-schedule futures, so many fewer futures than expected may
    actually be cancelled.
    """

    KILL_WORKERS = auto()
    """
    If the executor has a way of abruptly killing its workers, do so. Otherwise,
    this falls back to the `.CANCEL_FUTURES` behavior (e.g. for
    `~concurrent.futures.ThreadPoolExecutor`).

    .. note::

       This may skip `atexit` handlers or other normal worker cleanup. That
       said, this is frequently not a problem because:

        - executor-based applications typically need to be robust to abrupt
          worker failures anyway,

        - with this option, the whole worker pool is being shut down, and

        - the decoupling of workers from tasks in executor interfaces makes it
          harder to write the types of algorithms that require cross-process
          locks.
    """

    CANCEL = CANCEL_FUTURES
    """
    An alias for `.CANCEL_FUTURES`.
    """

    KILL = KILL_WORKERS
    """
    An alias for `.KILL_WORKERS`.
    """

    @classmethod
    def get(cls, arg: OnErrorLike) -> OnError:
        """
        Factory method that can take either a string or an instance of this
        class and return an instance to this class.
        """
        if isinstance(arg, str):
            return cls[arg.upper()]
        elif isinstance(arg, cls):
            return arg
        else:
            # Don't allow ints. This way we can safely use StrEnum in py 3.11+
            # without breaking users migrating from older versions of Python.
            raise TypeError(arg)


OnErrorLike = Union[OnError, str]
"""
The set of types accepted by `.OnError.get`.
"""


def extract_keywords(factory: Callable) -> Set[str]:
    """
    Helper for `.ExecutorPluginEntryPoint.allowed_keywords` properties. It
    inspects `.ExecutorPluginEntryPoint.create`-like functions an extracts their
    keyword argument names. This improves the maintainability of plugins where
    the factory publishes its accepted keywords in its call signature.

    :param factory: a function or method that creates an executor, takes at most
        one positional argument (other than possibly ``self``), and whose first
        (non-self) argument is called ``max_workers``.

    :return: the set of all keyword arguments the ``factory`` can accept, other
        than ``self`` and ``max_workers``. If ``factory`` is a class and its
        constructor takes a variadic ``**kwargs``-style argument, this function
        walks through its MRO superclasses to accumulate additional keywords,
        under the assumption that all subclass ``**kwargs`` get passed without
        name changes to the superclass.
    """
    kwarg_names = set()
    if isinstance(factory, type):
        funcs = factory.__mro__
    else:
        funcs = [factory]
    for func in funcs:
        if isinstance(func, type) and ("__init__" not in func.__dict__):
            # Skip mixins.
            continue
        saw_kwargs = False
        sig = signature(func)
        max_workers_idx = 0
        for i, (name, spec) in enumerate(sig.parameters.items()):
            if i == max_workers_idx:
                if spec.kind not in (
                    Parameter.POSITIONAL_ONLY,
                    Parameter.POSITIONAL_OR_KEYWORD,
                ):
                    raise TypeError(
                        f"The factory's first (non-self) parameter must "
                        f"be positional (and be for a max_workers argument).\n"
                        f"    factory:         {func}\n"
                        f"    parameter index: {i}\n"
                        f"    parameter name:  {name}\n"
                        f"    parameter kind:  {spec.kind}\n"
                    )
                if name == "self":
                    max_workers_idx = 1
                elif name == "max_workers":
                    pass
                else:
                    # For now, we'll be fussy about names for simplicity.
                    raise TypeError(
                        "The factory's first (non-self) parameter must "
                        "be named 'max_workers'."
                    )
            elif spec.kind == Parameter.POSITIONAL_ONLY:
                raise TypeError(
                    "The only allowed positional-only (non-self) parameter for "
                    "a factory is 'max_workers'."
                )
            elif spec.kind in (Parameter.POSITIONAL_OR_KEYWORD, Parameter.KEYWORD_ONLY):
                if spec.default is Parameter.empty:
                    raise TypeError(f"Parameter {name!r} has no default.")
                kwarg_names.add(name)
            elif spec.kind is Parameter.VAR_POSITIONAL:
                raise TypeError(f"Factories must not take *args-style varargs: {func}")
            elif spec.kind is Parameter.VAR_KEYWORD:
                # We've seen variadic kwargs, so if we're chasing the mro graph,
                # we need to keep going.
                saw_kwargs = True
            else:
                raise TypeError(f"Unknown factory parameter kind: {spec.kind}")
        if not saw_kwargs:
            break
    return kwarg_names


class ExecutorPluginEntryPoint(Protocol):
    """
    Each plugin module must have an ``ENTRY_POINT`` attribute that follows the
    protocol defined by this class. That variable tells the
    `pyrseus.ctx.mgr.ExecutorCtx` plugin system how to use the module's
    executor.
    """

    @property
    def supports_serial(self) -> bool:
        """
        Whether this plugin's ``ENTRY_POINT.create`` function can be called with
        ``max_workers`` = 0.
        """
        raise NotImplementedError

    @property
    def supports_concurrent(self) -> bool:
        """
        Whether this plugin's ``ENTRY_POINT.create`` function can be called with
        ``max_workers`` > 0.
        """
        raise NotImplementedError

    @property
    def is_available(self) -> bool:
        """
        Whether this plugin should be enabled. Typically this property will do
        an import check on the underlying executor module using
        `~pyrseus.core.sys.module_exists`.

        If this is false, `~pyrseus.ctx.mgr.ExecutorCtx` will never invoke the
        `.create` method for this plugin.
        """
        raise NotImplementedError

    @cached_property
    def allowed_keywords(self) -> Union[Tuple[str], List[str], Set[str]]:
        """
        The complete list of keywords that this plugin's ``ENTRY_POINT.create``
        method supports.

        If the user requests a `~pyrseus.ctx.mgr.ExecutorCtx` for this plugin,
        but they included some keywords that are missing from this set:

        - any keywords there *are* in any other available plugin's
          ``allowed_keywords`` will be silently dropped, and

        - any other keywords will trigger an exception.

        This filtering and validation is done by
        `~pyrseus.ctx.registry.filter_kwargs`.

        See the various plugins in the `pyrseus.ctx.plugins` package for
        examples of how to customize this property, especially when
        ``ENTRY_POINT.create`` takes a variadic ``**kwargs``-style argument.
        """
        return extract_keywords(self.create)

    def create(
        self,
        max_workers: Optional[int] = None,
        *,
        on_error: OnErrorLike = OnError.KILL_WORKERS,
        **kwargs,
    ) -> Tuple[Executor, Optional[PreExitFunc]]:
        """
        Creates an executor instance, returning both the instance and an
        optional function that will be called just before that executor's
        ``__exit__`` method is called at context exit time.

        To see the full list of keyword arguments that this plugin supports,
        use `~pyrseus.ctx.registry.get_keywords_for_plugin`.

        :param max_workers: the maximum number of workers the user has
            requested. ``ENTRY_POINT.create`` must support this as a positional
            argument, even for serial-only plugins that ignore its value.

        :param on_error: controls what `pyrseus.ctx.mgr.ExecutorCtx.__exit__`
            should do if an exception is passing through it. See `.OnError` for
            details.

        :param kwargs: additional arguments passed to the executor's
            constructor. These are filtered by
            `~pyrseus.ctx.registry.filter_kwargs`, using this plugin's
            ``ENTRY_POINT.allowed_keywords`` property.
        """
        raise NotImplementedError



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # `ExecutorCtx` vs. `get_executor`
#
# This notebook discusses how to use both `ExecutorCtx`
# and `get_executor` for the most common situations.

# %% [markdown]
# ## Setup
#
# Before running this notebook, make sure everything it depends on is installed:
#
#     # Modify the docs/requirements.txt path if you're running this
#     # command from anything except the repository's root directory.
#     python -m pip install -r docs/requirements.txt
#
# These are some imports that we'll use throughout this notebook:

# %% tags=["pyflyby-cell"]
import sys

sys.path.append("../../../src")  # assume we're running it from a Pyrseus source clone
import os
from concurrent.futures import as_completed
from contextlib import ExitStack

from pyrseus.ctx.mgr import ExecutorCtx
from pyrseus.interactive import get_executor

# %% [markdown]
# And here are some helper functions we'll be using throughout this notebook:


# %%
def get_row(r):
    """
    Returns the requested row from a 10x10 matrix with 0:10
    on the first row, 10:20 on the second row, etc.
    """
    if not (0 <= r < 10):
        raise IndexError(r)
    return list(range(r * 10, r * 10 + 10))


def get_col(c):
    """
    Returns the requested column from a 10x10 matrix with 0:10
    on the first row, 10:20 on the second row, etc.
    """
    if not (0 <= c < 10):
        raise IndexError(c)
    return list(range(c, 100, 10))


# %% [markdown]
# ## A Single Set of Tasks
#
# If your script or notebook has a single concurrent block, then `ExecutorCtx`
# can be a safe and convenient helper. Pick whatever plugin makes the most sense
# for your problem.
#
# If we want all futures to be attempted, even if some earlier ones fail, we can
# do this:

# %%
with ExecutorCtx("loky", 2) as exe:
    futs = [exe.submit(lambda r=r: sum(get_row(r))) for r in range(10)]
# With the fut.result() calls outside the with-statement, exe will keep
# trying to run every future, even if some fail.
print([fut.result() for fut in futs])

# %% [markdown]
# Or if we want to kill all workers immediately on the first detected error, we
# can check the results inside the with-statement. This optional feature is one
# of `ExecutorCtx`'s value adds (currently only supported with the `loky`
# plugin; see the `on_error` notebook for details):

# %%
with ExecutorCtx("loky", 2) as exe:
    futs = [exe.submit(lambda r=r: sum(get_row(r))) for r in range(10)]
    # With the fut.result() calls inside the with-statement, most
    # plugins will try to stop running additional futures if it can.
    print([fut.result() for fut in futs])

# %% [markdown]
# ## Multiple Sets of Tasks
#
# If your script or notebook needs to run multiple separate sets of tasks, there
# are several approaches to choose from.

# %% [markdown]
# ### ExecutorCtx with reusable loky executors
#
# Our `loky` plugin supports their reusable executors. Those executors try to
# reuse the existing worker pool from the previous time a reusable executor was
# requested. `ExecutorCtx` lets users combine `loky`'s reusable executors with
# the safety that context managers give us, including our choice of `on_error`
# handling (see the notebook of the same name for details).

# %%
# Do some work in a cell with an executor.
with ExecutorCtx("loky", 1, reusable=True) as exe:
    cell_1_worker_pid = exe.submit(os.getpid).result()
    futs = [exe.submit(lambda r=r: sum(get_row(r)), r) for r in range(10)]
    row_sums = [fut.result() for fut in futs]
    print(row_sums)
    print(sum(row_sums))

# %%
# And now we can do some more work in another cell, with another context
# manager.
with ExecutorCtx("loky", 1, reusable=True) as exe:
    cell_2_worker_pid = exe.submit(os.getpid).result()
    futs = [exe.submit(lambda c=c: sum(get_col(c)), c) for c in range(10)]
    col_sums = [fut.result() for fut in futs]
    print(col_sums)
    print(sum(col_sums))

# %%
# Both context managers used the same worker process.
assert cell_1_worker_pid == cell_2_worker_pid


# %% [markdown]
# ### Driver Functions that Take an Executor
#
# When writing library code, sometimes one has driver functions that want to
# run many concurrent tasks. If the author wants to support using an
# externally-supplied executor, they can have it passed in to the driver
# function or class instead of creating it themselves.
#
# Here's an example where we have two driver functions that we run with a single
# executor. This works with any executor, but it can requires work to avoid
# crosstalk between the drivers.


# %%
# This would typically be a library function.
def my_row_driver(exe):
    # Here we take an existing executor, submit tasks to it and track our tasks'
    # futures. We *don't* enter or exit the executor's context.
    #
    # Submit a bunch of tasks, including one that has a bad input (r=10).
    futs = []
    for r in range(11):
        futs.append(exe.submit(lambda r=r: sum(get_row(r)), r))
    # While we're at it, here we show a way to (a) use as_completed, (b) collate
    # results, and (c) save any task exceptions.
    #
    # This extra work is one of the downsides of this approach.
    row_sums = [None] * len(futs)
    for fut in as_completed(futs):
        r = futs.index(fut)
        try:
            row_sums[r] = fut.result()
        except Exception as ex:
            row_sums[r] = ex
    return row_sums


# This might be another library function.
def my_col_driver(exe):
    # As with my_row_driver, we'll intentionally submit one bad task (with c=10).
    futs = [exe.submit(lambda c=c: sum(get_col(c)), c) for c in range(11)]
    # Here we'll let the whole driver fail if there were any task errors.
    # But we'll at least be nice and cancel any remaining futures.
    #
    # This extra work is one of the downsides of this approach.
    col_sums = []
    for c in range(len(futs)):
        try:
            col_sums.append(futs[c].result())
        except Exception:
            # Try to cancel all remaining tasks. This is a noop for any
            # tasks that are already done.
            for c in range(c + 1, len(futs)):
                futs[c].cancel()
            # And reraise the exception
            raise
    return col_sums


# %%
# And this might be in our script or notebook where we want to run both drivers
# with an executor that we created and own.
try:
    with ExecutorCtx("thread", 1) as exe:
        print(f"row sums: {my_row_driver(exe)}")
        print(f"col sums: {my_col_driver(exe)}")
except Exception as ex:
    print(f"Caught {ex!r}")


# %% [markdown]
# ### Dual Use Driver Functions with ExitStack
#
# Another common situation is to have a driver function that can work in either
# the "A Single Set of Tasks" or the "Driver Functions that Take an Executor"
# way. The `ExitStack` context manager can make these dual-use driver functions
# easier to write.


# %%
# Here's a minimal example of a dual-use driver.
def my_dual_use_row_driver(exe=None):
    # First, we create a context manager that can be used to manage other
    # context managers, including none at all.
    with ExitStack() as es:
        # Now we conditionally create an executor and enter its context.
        if exe is None:
            exe = es.enter_context(ExecutorCtx("thread", 1))
        # And now we can proceed as usual.
        futs = [exe.submit(lambda r=r: sum(get_row(r)), r) for r in range(10)]
        # To make this safe, we'll cancel futures on error.
        #
        # This extra work is one of the downsides of this approach.
        try:
            return [fut.result() for fut in futs]
        except Exception:
            for fut in futs:
                fut.cancel()  # auto-ignored if it's already done
            raise


# %%
# Letting it create and manage its own executor.
my_dual_use_row_driver()

# %%
# Using our executor, managed by us.
with ExecutorCtx("cpprocess", 1) as exe:  # loky would be fine too
    print(my_dual_use_row_driver(exe=exe))

# %% [markdown]
# ### get_executor
#
# A final approach is to forego context managers completely and use
# `get_executor`. This saves the trouble of using a context manager, but it also
# removes all of the protections it gives, including `ExecutorCtx`'s
# enhancements to `loky`'s reusable executor.
#
# The benefits of using `get_executor` in Jupyter notebooks include:
#  - There's no need to create and use a context manager in each cell that
#    submits tasks.
#  - It can be convenient when using exiting "Driver Functions that Take an
#    Executor", as long as they're written properly to clean up after themselves
#    if they raise an exception.
#  - It allows reusing the same executor across many cells, for any plugin, not
#    just the `loky` one.
#  - It helps bypass long startup times for some plugins like `ipyparallel`.
#
# The downsides include:
#  - Users must remember to block on every future instead of letting the context
#    manager do that automatically and implicitly at ``__exit__`` time.
#  - If a cell that's managing futures is interrupted, submitted tasks will
#    continue to run in the background, potentially creating a silent backlog
#    for new tasks submitted afterward.
#  - Finalizers are not reliable for resource management in Python, so the
#    wrapper's destructor may never be triggered. This can lead to resource
#    leaks: tasks continuing to run when the user didn't intend them to, workers
#    not being torn down, worker management threads or processes not being torn
#    down, etc.
#
# Here is an example of using it with the dual-use driver from the previous
# section:

# %%
print(my_dual_use_row_driver(exe=get_executor("cpprocess", 1)))



"""
Verifies that `.ExecutorCtx` creates exactly the right number of workers with
the ``"thread"`` and ``"process"`` plugins.
"""

import os
import threading
from contextlib import ExitStack
from multiprocessing import Manager

import pytest

from pyrseus.core.sys import get_num_available_cores
from pyrseus.ctx.mgr import ExecutorCtx
from pyrseus.ctx.registry import skip_if_unavailable

PROC_TIMEOUT_SECS = 1  # should be long enough
THREAD_TIMEOUT_SECS = 0.1  # should be long enough


def coordinated_get_pid_and_tid(parent_pid, parent_tid, barrier, concurrency_style):
    """
    Test helper that ensures that our tasks are all assigned to different
    workers.

    This function blocks until all tasks are also waiting on the `barrier`, then
    fetches the current pid and tid and returns them.
    """
    # Also verify that we're
    if concurrency_style == "serial":
        assert os.getpid() == parent_pid
        assert threading.get_ident() == parent_tid
    elif concurrency_style == "thread":
        assert os.getpid() == parent_pid
        assert threading.get_ident() != parent_tid
    elif concurrency_style == "process":
        assert os.getpid() != parent_pid
    else:
        # If we ever add cross-host testing, then check some relevant host
        # identifier too like the primary IP address.
        raise ValueError(concurrency_style)
    # Wait for all workers to be running this function. This is an easy way to
    # verify how many workers have been started.
    barrier.wait()
    # Tell the main process what this worker's process and thread ids are.
    return os.getpid(), threading.get_ident()


@pytest.mark.parametrize("max_workers", (None, 0, 1, 2))
@pytest.mark.parametrize(
    "plugin,concurrency_style",
    (
        ("cpnocatch", "serial"),
        ("cpinline", "serial"),
        ("cpmpi4py", "process"),
        ("cpprocess", "process"),
        ("nocatch", "serial"),
        ("inline", "serial"),
        ("ipyparallel", "process"),
        ("loky", "process"),
        ("mpi4py", "process"),
        ("pnocatch", "serial"),
        ("pinline", "serial"),
        ("process", "process"),
        ("thread", "thread"),
    ),
)
def test_plugin_pool_size(plugin, concurrency_style, max_workers):
    skip_if_unavailable(plugin)

    if max_workers is None:
        ctx_args = ()
        if concurrency_style == "serial":
            num_tasks = 1
        else:
            # This is the current default for all of the built-in concurrent
            # plugins, except for the mpi4py ones.
            num_tasks = max(1, get_num_available_cores())
    else:
        ctx_args = (max_workers,)
        num_tasks = max(1, max_workers)

    if "mpi4py" in plugin:
        # Problem 1: in its default configuration, mpi4py likes to set slots=1,
        # and it's inconvenient to either override or detect this.
        #
        # Problem 2: mpi4py doesn't play nicely with
        # multiprocessing.Manager.Barrier, at least right out of the box. In
        # particular, it takes some work to securely setup a shared
        # multiprocessing authentication secret. A real MPI user would use
        # MPI-specific mechanisms anyway.
        pytest.xfail("This test is too hard to setup with mpi4py.")
    elif "ipyparallel" in plugin:
        # Problem 1 does *not* apply to ipyparallel. It's fine with
        # oversubscribing CPU resources.
        #
        # Problem 2 also applies to ipyparallel.
        pytest.xfail("This test is too hard to setup with ipyparallel.")

    if concurrency_style == "serial":
        timeout_secs = 1e-9
        if num_tasks > 1:
            pytest.xfail(
                "For serial plugins, this test is only applicable for <2 workers."
            )
    elif concurrency_style == "thread":
        timeout_secs = THREAD_TIMEOUT_SECS
    elif concurrency_style == "process":
        timeout_secs = PROC_TIMEOUT_SECS
    else:
        raise ValueError(concurrency_style)

    parent_pid = os.getpid()
    parent_tid = threading.get_ident()
    with ExitStack() as es:
        # Setup the manager that will create the Barrier object for us.
        sync_mgr = es.enter_context(Manager())

        # Empty for now, but useful to keep for troubleshooting.
        ctx_kwargs = {}

        # Create the ExecutorCtx. It will auto-adjust max_workers if it's zero
        # and the plugin is concurrent-only, or if it's non-zero and the plugin
        # is serial-only.
        ctx = ExecutorCtx(plugin, *ctx_args, **ctx_kwargs)
        exe = es.enter_context(ctx)

        # First check that at least the expected number of workers are started.
        futs = []
        barrier = sync_mgr.Barrier(num_tasks, timeout=timeout_secs)
        for _ in range(num_tasks):
            futs.append(
                exe.submit(
                    coordinated_get_pid_and_tid,
                    parent_pid,
                    parent_tid,
                    barrier,
                    concurrency_style,
                )
            )
        assert len({fut.result() for fut in futs}) == num_tasks

        # Next, verify that no extra concurrent workers were started by making a
        # barrier that's too wide.
        #
        # This is the same test as above, except (a) we use num_tasks+1 when
        # setting up the barrier, and (b) we expect every task to fail instead
        # of none of them.
        futs = []
        barrier = sync_mgr.Barrier(num_tasks + 1, timeout=timeout_secs)
        for _ in range(num_tasks + 1):
            with ExitStack() as es2:
                if plugin.endswith("nocatch"):
                    es2.enter_context(pytest.raises(threading.BrokenBarrierError))
                futs.append(
                    exe.submit(
                        coordinated_get_pid_and_tid,
                        parent_pid,
                        parent_tid,
                        barrier,
                        concurrency_style,
                    )
                )
        for fut in futs:
            # Every future should fail.
            with pytest.raises(threading.BrokenBarrierError):
                fut.result()



"""
Manages `~pyrseus.ctx.mgr.ExecutorCtx`'s plugin registry.

`~pyrseus.ctx.mgr.ExecutorCtx` is a factory class that can create many different
types of `~concurrent.futures.Executor`s. This module provides the mechanism it
uses to track all of those different types, via a registry of named
`pyrseus.ctx.api.ExecutorPluginEntryPoint` objects.

See :ref:`writingplugins` for more information on how to create a plugin to wrap
your own executor classes.

Related Modules and Packages
----------------------------

- `pyrseus.executors`: a package providing some but not all of the underlying
  executors that are made available to `~pyrseus.ctx.mgr.ExecutorCtx` via this
  registry. Users are free to directly use those executor classes if they do not
  wish to use `~pyrseus.ctx.mgr.ExecutorCtx`.

- `pyrseus.ctx.api`: provides the base class
  `~pyrseus.ctx.api.ExecutorPluginEntryPoint`. This module maintains a registry
  of concrete instances of that class.

- `pyrseus.ctx.mgr`: provides `~pyrseus.ctx.mgr.ExecutorCtx`, the primary client
  of this module.

- `pyrseus.ctx.plugins`: a package of plugin modules for
  `~pyrseus.ctx.mgr.ExecutorCtx`. Each module in that package must have a global
  variable called ``ENTRY_POINT`` that is an
  `~pyrseus.ctx.api.ExecutorPluginEntryPoint` instance. The plugin's name will
  be the same as the module's name, after ``"pyrseus.ctx.plugins."``. This
  registry auto-discovers all such entry points at import time.

- `pyrseus.ctx.simple`: provides a simpler alternative API to the base one
  provided by `pyrseus.ctx.api`. Plugin authors can use whichever API suits
  their needs better.

Handling Keywords
-----------------

The following functions are related to validating and filtering keyword
arguments so that it can both (a) silently ignore irrelevant keywords for a
particular plugin, yet (b) still provide useful errors for keywords that aren't
valid for any available plugin:

- `filter_kwargs`
- `get_all_allowed_keywords`
- `get_keywords_by_plugin`
- `get_keywords_for_plugin`

Test Helpers
------------

The following functions are primarily useful for writing tests that only work if
a particular plugin is installed:

- `is_plugin_available`
- `skip_if_unavailable`

Registry Management
-------------------

The following are useful for managing the registry itself.

- `get_entry_point`: retrieves the requested plugin's
  `~pyrseus.ctx.api.ExecutorPluginEntryPoint` object, if possible. This is what
  `~pyrseus.ctx.mgr.ExecutorCtx` uses.

- `register_plugin`: registers a new or replacement plugin. This is called
  automatically for each module in the `pyrseus.ctx.plugins` package when this
  registry module is imported. It can also be directly called for 3rd party
  plugins.

- `SetDefaultExecutorPluginCtx`: temporarily changes the default concurrent
  and/or serial plugin, with standard context manager semantics.

- `set_default_executor_plugin`: permanently changes the default concurrent
  and/or serial plugin.
"""

import importlib
import pkgutil
from contextlib import contextmanager
from typing import Any, Dict, Optional, Set, Tuple, Union

from .api import ExecutorPluginEntryPoint

__ALL__ = [
    "filter_kwargs",
    "get_all_allowed_keywords",
    "get_entry_point",
    "get_keywords_by_plugin",
    "get_keywords_for_plugin",
    "is_plugin_available",
    "register_plugin",
    "set_default_executor_plugin",
    "SetDefaultExecutorPluginCtx",
    "skip_if_unavailable",
]

_SERIAL_ENTRY_POINTS: Dict[str, ExecutorPluginEntryPoint] = {}
_CONCURRENT_ENTRY_POINTS: Dict[str, ExecutorPluginEntryPoint] = {}
_KNOWN_BUT_UNAVAILABLE_ENTRY_POINTS: Dict[str, ExecutorPluginEntryPoint] = {}
_DEFAULT_SERIAL_ENTRY_POINT_NAME: str = "inline"
_DEFAULT_CONCURRENT_ENTRY_POINT_NAME: str = "cpprocess"
_ALL_AVAILABLE_KEYWORDS: Set[str] = set()


def get_all_allowed_keywords():
    """
    Returns the full set of allowed keyword arguments to
    `~pyrseus.ctx.mgr.ExecutorCtx`. This is the union of the allowed keyword
    arguments across all registered plugins.

    >>> get_all_allowed_keywords()
    {...'mp_context'...}
    """
    return _ALL_AVAILABLE_KEYWORDS


def get_keywords_for_plugin(
    plugin_or_max_workers: Optional[Union[str, int]] = None,
    /,
    max_workers: Optional[int] = None,
):
    """
    Returns the full set of keyword arguments that will be forwarded to a
    particular plugin.

    >>> get_keywords_for_plugin("process")
    {...'mp_context'...}

    >>> assert get_keywords_for_plugin("process") <= get_all_allowed_keywords()
    """
    plugin, max_workers = _interpret_plugin_and_max_workers(
        plugin_or_max_workers, max_workers
    )
    _, _, entry_point = get_entry_point(plugin, max_workers)
    return entry_point.allowed_keywords


def get_keywords_by_plugin():
    """
    Returns the full set of keyword arguments, on a per-plugin basis.

    >>> by_plugin = get_keywords_by_plugin()
    >>> sorted(by_plugin['inline'])
    []
    >>> sorted(by_plugin['process'])
    [...'initializer'...'mp_context'...]

    """
    ret = {}
    for mapping in (_CONCURRENT_ENTRY_POINTS, _SERIAL_ENTRY_POINTS):
        for plugin, entry_point in mapping.items():
            accum = ret.setdefault(plugin, set())
            accum |= entry_point.allowed_keywords
    return ret


def register_plugin(name: str, entry_point: ExecutorPluginEntryPoint) -> bool:
    """
    Registers a new `.ExecutorPluginEntryPoint` as a plugin associated with the
    given ``name``.

    Registering an unavailable plugin is fine: this module will respect
    `entry_point.is_available <.ExecutorPluginEntryPoint.is_available>` whenever
    it uses the entry point object.
    """
    if not entry_point.is_available:
        # Silently ignore any plugins that don't have their requirements met or
        # have otherwise been disabled.
        _KNOWN_BUT_UNAVAILABLE_ENTRY_POINTS[name] = entry_point
    else:
        # Extract and validate the metadata before mutating and global data
        # structures.
        allowed_keywords = entry_point.allowed_keywords
        supports_serial = entry_point.supports_serial
        supports_concurrent = entry_point.supports_concurrent
        if not isinstance(allowed_keywords, (tuple, list, set)):
            raise TypeError(f"{entry_point.allowed_keywords=} must be a set-like")
        if not isinstance(supports_serial, bool):
            raise TypeError(f"{entry_point.supports_serial=} must be a bool")
        if not isinstance(supports_concurrent, bool):
            raise TypeError(f"{entry_point.supports_concurrent=} must be a bool")
        if not supports_serial and not supports_concurrent:
            raise RuntimeError(
                f"The plugin for {name} must support at least one mode "
                f"(serial and/or concurrent)."
            )
        # Perform the registration.
        _ALL_AVAILABLE_KEYWORDS.update(allowed_keywords)
        _KNOWN_BUT_UNAVAILABLE_ENTRY_POINTS.pop(name, None)
        if supports_serial:
            _SERIAL_ENTRY_POINTS[name] = entry_point
        if supports_concurrent:
            _CONCURRENT_ENTRY_POINTS[name] = entry_point


def set_default_executor_plugin(
    name: str, serial: Optional[bool] = None, concurrent: Optional[bool] = None
) -> Tuple[str, str]:
    """
    Permanently sets the default serial and/or concurrent plugin for the current
    process. This is primarily intended for:

    - being the backend for `.SetDefaultExecutorPluginCtx`,
    - setting the defaults for entire interactive Jupyter notebook and IPython
      sessions,
    - setting the defaults for entire scripts, and
    - configuring user- and site-wide defaults.
    """
    global _DEFAULT_SERIAL_ENTRY_POINT_NAME
    global _DEFAULT_CONCURRENT_ENTRY_POINT_NAME
    old_serial = _DEFAULT_SERIAL_ENTRY_POINT_NAME
    old_concurrent = _DEFAULT_CONCURRENT_ENTRY_POINT_NAME

    if serial is None:
        if name in _SERIAL_ENTRY_POINTS:
            _DEFAULT_SERIAL_ENTRY_POINT_NAME = name
    elif serial:
        if name not in _SERIAL_ENTRY_POINTS:
            raise ValueError(f"{name} is a concurrent-only plugin")
        _DEFAULT_SERIAL_ENTRY_POINT_NAME = name

    if concurrent is None:
        if name in _CONCURRENT_ENTRY_POINTS:
            _DEFAULT_CONCURRENT_ENTRY_POINT_NAME = name
    elif concurrent:
        if name not in _CONCURRENT_ENTRY_POINTS:
            raise ValueError(f"{name} is a serial-only plugin")
        _DEFAULT_CONCURRENT_ENTRY_POINT_NAME = name

    return old_serial, old_concurrent


@contextmanager
def SetDefaultExecutorPluginCtx(
    name: str, serial: Optional[bool] = None, concurrent: Optional[bool] = None
):
    """
    Temporarily sets the default serial and/or concurrent plugin.

    >>> print("Default serial entry point:    ", get_entry_point(0)[0])
    Default serial entry point:     inline
    >>> print("Default concurrent entry point:", get_entry_point(1)[0])
    Default concurrent entry point: cpprocess

    >>> skip_if_unavailable("cpinline")
    >>> with SetDefaultExecutorPluginCtx("cpinline"):
    ...     print("Default serial entry point:    ", get_entry_point(0)[0])
    ...     print("Default concurrent entry point:", get_entry_point(1)[0])
    Default serial entry point:     cpinline
    Default concurrent entry point: cpprocess

    >>> with SetDefaultExecutorPluginCtx("process"):
    ...     print("Default serial entry point:    ", get_entry_point(0)[0])
    ...     print("Default concurrent entry point:", get_entry_point(1)[0])
    Default serial entry point:     inline
    Default concurrent entry point: process

    This function guards against mixing up concurrent vs. serial plugins.

    >>> skip_if_unavailable("cpinline")
    >>> with SetDefaultExecutorPluginCtx("cpinline", concurrent=True):
    ...     pass
    Traceback (most recent call last):
    ...
    ValueError: cpinline is a serial-only plugin
    """
    global _DEFAULT_SERIAL_ENTRY_POINT_NAME
    global _DEFAULT_CONCURRENT_ENTRY_POINT_NAME
    old_serial, old_concurrent = set_default_executor_plugin(name, serial, concurrent)
    try:
        yield old_serial, old_concurrent
    finally:
        _DEFAULT_SERIAL_ENTRY_POINT_NAME = old_serial
        _DEFAULT_CONCURRENT_ENTRY_POINT_NAME = old_concurrent


def get_entry_point(
    plugin_or_max_workers: Optional[Union[str, int]] = None,
    /,
    max_workers: Optional[int] = None,
) -> Tuple[str, Optional[int], ExecutorPluginEntryPoint]:
    """
    Returns the resolved plugin ``name``, (potentially resolved)
    ``max_workers``, and `.ExecutorPluginEntryPoint` object associated with the
    requested plugin.

    >>> get_entry_point("inline")[2].allowed_keywords
    set()

    >>> get_entry_point("not-a-plugin")
    Traceback (most recent call last):
    ...
    ValueError: Unknown ExecutorCtx plugin name: 'not-a-plugin'

    :param name: the name of the desired plugin. If `None`, then ``max_workers``
        will be used to auto-choose between the default serial and default
        concurrent plugin.

    :param max_workers: used as needed to help find the right plugin.
    """
    name, max_workers = _interpret_plugin_and_max_workers(
        plugin_or_max_workers, max_workers
    )
    if max_workers is None:
        if name is None:
            name = _DEFAULT_CONCURRENT_ENTRY_POINT_NAME
            entry_point = _CONCURRENT_ENTRY_POINTS[name]
        elif name in _CONCURRENT_ENTRY_POINTS:
            entry_point = _CONCURRENT_ENTRY_POINTS[name]
        elif name in _SERIAL_ENTRY_POINTS:
            entry_point = _SERIAL_ENTRY_POINTS[name]
        elif name in _KNOWN_BUT_UNAVAILABLE_ENTRY_POINTS:
            raise ValueError(
                f"There is a {name!r} ExecutorCtx plugin, but the plugin "
                f"has marked itself as unavailable in this process."
            )
        else:
            raise ValueError(f"Unknown ExecutorCtx plugin name: {name!r}")
    elif name is None:
        if max_workers == 0:
            name = _DEFAULT_SERIAL_ENTRY_POINT_NAME
            entry_point = _SERIAL_ENTRY_POINTS[name]
        else:
            # We accept any non-zero max_workers here, not just positive ones.
            # -1 often means, "pick a good max for me."
            name = _DEFAULT_CONCURRENT_ENTRY_POINT_NAME
            entry_point = _CONCURRENT_ENTRY_POINTS[name]
    else:
        if name in _KNOWN_BUT_UNAVAILABLE_ENTRY_POINTS:
            raise ValueError(
                f"There is a {name!r} ExecutorCtx plugin, but the plugin "
                f"has marked itself as unavailable in this process. Has its "
                f"dependencies been installed?"
            )
        elif max_workers == 0:
            # Be nice and auto-switch to a concurrent one if needed.
            if name in _SERIAL_ENTRY_POINTS:
                entry_point = _SERIAL_ENTRY_POINTS[name]
            elif name in _CONCURRENT_ENTRY_POINTS:
                entry_point = _CONCURRENT_ENTRY_POINTS[name]
                # We only get here for concurrent-only entry points, and those
                # typically raise exceptions if max_workers==0. So silently
                # upgrade to max_workers=1.
                max_workers = 1
            else:
                raise ValueError(
                    f"Unknown serial ExecutorCtx plugin name: {name!r}\n"
                    f"Known names: {sorted(_SERIAL_ENTRY_POINTS)}"
                )
        else:
            # Be nice an auto-switch to a serial one if needed.
            if name in _CONCURRENT_ENTRY_POINTS:
                entry_point = _CONCURRENT_ENTRY_POINTS[name]
            elif name in _SERIAL_ENTRY_POINTS:
                entry_point = _SERIAL_ENTRY_POINTS[name]
                # We only get here for serial-only entry points. TO be
                # consistent with the behavior when auto-switching to
                # concurrent-only entry points, we'll override max_workers.
                max_workers = 0
            else:
                raise ValueError(
                    f"Unknown concurrent ExecutorCtx plugin name: {name!r}.\n"
                    f"Known names: {sorted(_CONCURRENT_ENTRY_POINTS)}"
                )

    return name, max_workers, entry_point


def is_plugin_available(name: str) -> bool:
    """
    Returns whether the given plugin is known and is available.

    >>> is_plugin_available("inline")
    True
    >>> is_plugin_available("not-a-plugin")
    False
    """
    return (name not in _KNOWN_BUT_UNAVAILABLE_ENTRY_POINTS) and (
        (name in _SERIAL_ENTRY_POINTS) or (name in _CONCURRENT_ENTRY_POINTS)
    )


def skip_if_unavailable(name: str):
    """
    Unit test helper that skips the current doctest or formal unit test if the
    given plugin is not available in the current Python installation.
    """
    if not is_plugin_available(name):
        import pytest

        all_plugin_names = sorted(
            set(_CONCURRENT_ENTRY_POINTS) | set(_SERIAL_ENTRY_POINTS)
        )
        known_plugin_names = sorted(_KNOWN_BUT_UNAVAILABLE_ENTRY_POINTS)
        # We need to use pytest.skip and not unittest.SkipTest: the latter
        # breaks doctests, at least for pytest.
        pytest.skip(
            f"The {name!r} plugin is not available.\n"
            f"    available: {all_plugin_names}\n"
            f"    known but unavailable: {known_plugin_names}"
        )


def _interpret_plugin_and_max_workers(
    plugin_or_max_workers: Optional[Union[str, int]] = None,
    /,
    max_workers: Optional[int] = None,
) -> Tuple[str, Optional[int]]:
    """
    Helper for `.get_entry_point` that decodes the first 0-2 positional
    arguments expected by `~pyrseus.ctx.mgr.ExecutorCtx`'s constructor, but
    without inspecting the plugin registry. See the constructor's documentation
    for an in-depth discussion.

    >>> _interpret_plugin_and_max_workers()
    (None, None)
    >>> _interpret_plugin_and_max_workers(0)
    (None, 0)
    >>> _interpret_plugin_and_max_workers(4)
    (None, 4)
    >>> _interpret_plugin_and_max_workers("process")
    ('process', None)
    >>> _interpret_plugin_and_max_workers("process", 4)
    ('process', 4)

    """
    if plugin_or_max_workers is None:
        if max_workers is not None:
            raise ValueError(
                f"{max_workers=} should only be supplied when the "
                f"first argument is a string."
            )
        plugin_name = None
    elif max_workers is None:
        if isinstance(plugin_or_max_workers, str):
            plugin_name = plugin_or_max_workers
        else:
            plugin_name = None
            max_workers = int(plugin_or_max_workers)
    else:
        plugin_name = plugin_or_max_workers

    return plugin_name, max_workers


def filter_kwargs(
    entry_point: ExecutorPluginEntryPoint, kwargs: Dict[str, Any]
) -> Dict[str, Any]:
    """
    This function performs the keyword validation and filtering for
    `~pyrseus.ctx.mgr.ExecutorCtx`.

    It removes keyword arguments from ``kwargs`` that are not accepted by
    ``entry_point``'s `create <.ExecutorPluginEntryPoint.create>` method. To
    help avoid hiding too many errors, this function also raises an exception if
    ``kwargs`` has any keywords that are now allowed by any known entry points.

    Keywords that are supported by the given entry point are preserved.

        >>> _, _, process_ep = get_entry_point("process", 1)
        >>> filter_kwargs(process_ep, {"initializer": None})
        {'initializer': None}

    Keywords that are not supported by any registered entry point result in
    exceptions.

        >>> filter_kwargs(process_ep, {"not-a-keyword-for-any-plugin": 42})
        Traceback (most recent call last):
        ...
        TypeError: Keyword "not-a-keyword-for-any-plugin" is not accepted ...

    Keywords that are not supported by the given entry point but are supported
    by another are silently dropped.

        >>> skip_if_unavailable("loky")
        >>> _, _, loky_ep = get_entry_point("loky", 1)
        >>> filter_kwargs(process_ep, {"reuse": True, "initializer": None})
        {'initializer': None}
        >>> filter_kwargs(loky_ep, {"reuse": True, "initializer": None})
        {'reuse': True, 'initializer': None}
    """
    filtered_kwargs = {}
    for k, v in kwargs.items():
        if k not in _ALL_AVAILABLE_KEYWORDS:
            all_plugin_names = sorted(
                set(_CONCURRENT_ENTRY_POINTS) | set(_SERIAL_ENTRY_POINTS)
            )
            raise TypeError(
                f"Keyword {k!r} is not accepted by any of the registered plugins:\n"
                f"    plugins: {all_plugin_names}\n"
                f"    all allowed keywords: {sorted(_ALL_AVAILABLE_KEYWORDS)}"
            )
        elif k in entry_point.allowed_keywords:
            filtered_kwargs[k] = v
        else:
            # Silently drop those that can be used by other plugins but aren't
            # used by the currently-selected plugin.
            pass

    return filtered_kwargs


def _register_plugins_from_package():
    """
    Helper used at import time to pre-register all plugins found in the
    `pyrseus.ctx.plugins` package.

    Future work: if there's sufficient user demand, add support for either the
    "Using naming convention" or "Using package metadata" approach to
    auto-discovering 3rd party plugins, as described here:
    https://packaging.python.org/en/latest/guides/creating-and-discovering-plugins/.
    We previously experimented with using the "Using namespace packages"
    approach, but found it to be too limiting and cumbersome.
    """
    from pyrseus.ctx import plugins

    # This code actually implements the full "Using namespace packages"
    # approach. But to make it work, we'd need to change all of pyrseus,
    # pyrseus.ctx, and pyrseus.ctx.plugins to namespace packages. This would
    # prevent us from providing convenient transplanted imports in the top-level
    # pyrseus module, and prevent us from embedding __version__ there too.
    search_path = plugins.__path__
    search_prefix = plugins.__name__ + "."
    fqns = [fqn for _, fqn, _ in pkgutil.iter_modules(search_path, search_prefix)]
    for fqn in sorted(fqns):  # sort for better repeatability
        mod = importlib.import_module(fqn)
        _, plugin_name = fqn.rsplit(".", 1)
        if not hasattr(mod, "ENTRY_POINT"):
            # For now, we'll be strict and insist that there's no cruft in the
            # pyrseus.ctx.plugins package.
            raise RuntimeError(
                f"The plugin module {fqn} at {mod.__file__} does "
                f"not have an ENTRY_POINT attribute."
            )
        try:
            register_plugin(plugin_name, mod.ENTRY_POINT)
        except Exception as ex:
            raise RuntimeError(
                f"Failed to register the plugin module {fqn} at {mod.__file__}."
            ) from ex


_register_plugins_from_package()

# Right now, we hard-code the initial defaults and the point to ones that ship
# with this package, so they should always be available. We can adjust these
# rules later if there's sufficient user demand.
if _DEFAULT_SERIAL_ENTRY_POINT_NAME not in _SERIAL_ENTRY_POINTS:
    raise ImportError(
        f"The initial default serial plugin, "
        f"{_DEFAULT_SERIAL_ENTRY_POINT_NAME!r} has not been installed. "
        f"All known serial plugins: "
        f"{sorted(_SERIAL_ENTRY_POINTS)}"
    )
if _DEFAULT_CONCURRENT_ENTRY_POINT_NAME not in _CONCURRENT_ENTRY_POINTS:
    raise ImportError(
        f"The initial default concurrent plugin, "
        f"{_DEFAULT_CONCURRENT_ENTRY_POINT_NAME!r} has not been installed. "
        f"All known serial plugins: "
        f"{sorted(_CONCURRENT_ENTRY_POINTS)}"
    )



# Configuration file for the Sphinx documentation builder.

import sys
import time

from jupytext.cli import jupytext
from sphinx.ext import autodoc

# -- Tell sphinx where to find all of our code and tests.

# Insert this clone's copy of Pyrseus first, just in case there is another,
# older version installed.
sys.path[:] = ["../../src", "../../tests/"] + sys.path

# -- Hack to create paired notebook files so that myst_nb can find them.

jupytext(("--update", "--to", "ipynb", "notebooks/*.py"))

# -- Project information

# If this import fails, rerun configure.py at top of the repo.
from pyrseus import __version__ as version_str  # NOQA

project = "pyrseus"
copyright = f"2023-{time.strftime('%Y')}, Gerald Dalley"
author = "Gerald Dalley"

release = ".".join(version_str.split(".")[:2])
version = version_str

# -- General configuration

extensions = [
    "myst_nb",
    "sphinx_design",
    "sphinx_inline_tabs",
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.duration",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
]
autosummary_generate = True

intersphinx_disabled_domains = ["std"]

templates_path = ["_templates"]

highlight_language = "python3"

default_role = "autolink"

sphinx_tabs_disable_tab_closing = True
sphinx_tabs_disable_css_loading = True

# Hack from https://stackoverflow.com/a/75041544 to suppress bloated docs about
# things inherited from the base `object` type.


class MockedClassDocumenter(autodoc.ClassDocumenter):
    def add_line(self, line: str, source: str, *lineno: int) -> None:
        if line == "   Bases: :py:class:`object`":
            return
        super().add_line(line, source, *lineno)


autodoc.ClassDocumenter = MockedClassDocumenter

# For cross references to other packages we interface with.
intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "sphinx": ("https://www.sphinx-doc.org/en/master/", None),
    "dill": ("https://dill.readthedocs.io/en/latest", None),
    "ipyparallel": ("https://ipyparallel.readthedocs.io/en/latest", None),
    "loky": ("https://loky.readthedocs.io/en/stable", None),
    "mpi4py": ("https://mpi4py.readthedocs.io/en/stable", None),
}

# -- Options for HTML output

html_theme = "furo"
html_static_path = ["_static"]
html_css_files = [
    "css/custom.css",
]

rst_epilog = """
.. |nbh| unicode:: U+2011
   :trim:
.. |nbsp| unicode:: U+00A0
   :trim:
.. |br| raw:: html

   <br/>
.. |nobr| raw:: html

   <nobr>
.. |nobrc| raw:: html

   </nobr>

.. |cloudpickle| replace:: cloudpickle
.. _cloudpickle: https://github.com/cloudpipe/cloudpickle

.. |dill| replace:: dill
.. _dill: https://dill.readthedocs.io

.. |ipyparallel| replace:: ipyparallel
.. _ipyparallel: https://ipyparallel.readthedocs.io

.. |loky| replace:: loky
.. _loky: https://loky.readthedocs.io

.. |mpi4py| replace:: mpi4py
.. _mpi4py: https://mpi4py.readthedocs.io

.. |psutil| replace:: psutil
.. _psutil: https://psutil.readthedocs.io

.. |nb_get_executor| replace:: ExecutorCtx vs. get_executor
.. _nb_get_executor: ../notebooks/get_executor.html

"""

# -- Options for EPUB output
epub_show_urls = "footnote"



"""
Provides a drop-in replacement for the built-in `.ProcessPoolExecutor` that uses
|cloudpickle|_ for pickling tasks and their return values, instead of `pickle`.
"""

from concurrent.futures import ProcessPoolExecutor

import cloudpickle
from pyrseus.core.pickle import CustomPickledClosure


class CpProcessPoolExecutor(ProcessPoolExecutor):
    """
    A drop-in replacement for the built-in `.ProcessPoolExecutor` that uses
    |cloudpickle|_ for pickling tasks and their return values, instead of
    `pickle`.

    Consider this lambda.

        >>> needs_cloudpickle = lambda: 123

    It can't be pickled with `pickle`.

        >>> import pickle
        >>> pickle.dumps(needs_cloudpickle, -1)
        Traceback (most recent call last):
        ...
        _pickle.PicklingError: Can't pickle ...

    And since the built-in `.ProcessPoolExecutor` uses `pickle`, it can't handle
    it:

        >>> with ProcessPoolExecutor(1) as exe:
        ...     fut = exe.submit(needs_cloudpickle)
        ...     print(fut.result())
        Traceback (most recent call last):
        ...
        _pickle.PicklingError: Can't pickle ...

    But `.CpProcessPoolExecutor` is a thin wrapper around `.ProcessPoolExecutor`
    that uses |cloudpickle|_ for serialization, so it works fine.

        >>> with CpProcessPoolExecutor(1) as exe:
        ...     fut = exe.submit(needs_cloudpickle)
        ...     print(fut.result())
        123

    See :doc:`../plugins` for a list of related executors.
    """

    def submit(self, fcn, /, *args, **kwargs):
        closure = CustomPickledClosure(
            fcn,
            args,
            kwargs,
            dumps=cloudpickle.dumps,
            loads=cloudpickle.loads,
        )
        return super().submit(closure)



"""
`~pyrseus.ctx.mgr.ExecutorCtx` plugin that uses `.loky`'s enhanced
multiprocessing executors.

Some imports for our tests:

    >>> import os, time
    >>> from pyrseus.ctx.registry import skip_if_unavailable
    >>> from pyrseus.ctx.mgr import ExecutorCtx
    >>> skip_if_unavailable("loky")
    >>> from loky import get_reusable_executor

|loky|_'s executors can be used as an improved backend to the ``"process"`` and
``"cpprocess"`` plugins that use the built-in
`~concurrent.futures.ProcessPoolExecutor`.

    >>> # "loky" can pickle fancy things like lambdas with cloudpickle
    >>> skip_if_unavailable("loky")
    >>> with ExecutorCtx("loky", 1) as exe:
    ...     assert exe.submit(lambda: os.getpid()).result() != os.getpid()

Additionally, the ``"loky"`` plugin allows using |loky|_'s reusable executors,
but with the safety benefits of a context manager. The manager's exit waits till
all tasks submitted via it enter a final state. Any tasks submitted to the
reusable executor through any other means are untouched.

    >>> skip_if_unavailable("loky")
    >>> # Note: protected_exe and unprotected_exe will use the same underlying
    >>> # loky reusable executor.
    >>> unprotected_exe = get_reusable_executor(1)
    >>> with ExecutorCtx("loky", 1, reusable=True) as protected_exe:
    ...     # Submit via the protected context manager.
    ...     # protected_exe.__exit__ will wait for this future to be completed.
    ...     protected_fut = protected_exe.submit(os.getpid)
    ...
    ...     # Submit another, but via the unprotected interface.
    ...     # protected_exe.__exit__ ignores this future.
    ...     unprotected_fut = unprotected_exe.submit(
    ...         # loky handles lambdas fine
    ...         lambda: time.sleep(0.2) or os.getpid()
    ...     )
    >>> # All futures submitted to the ExecutorCtx are completed before exiting.
    >>> assert protected_fut.done()
    >>> # But futures submitted directly to the same underlying reusable
    >>> # executor are unaffected by this. Here we see that unprotected_fut
    >>> # is still sleeping.
    >>> assert not unprotected_fut.done()
    >>> # If we explicitly wait till unprotected_fut is completed too, we can
    >>> # see that it ran on the same worker as protected_fut.
    >>> assert unprotected_fut.result() == protected_fut.result()
    >>> # Clean up must be done explicitly with reusable executors.
    >>> unprotected_exe.shutdown(kill_workers=True)

See `.EntryPoint.create` for more details.

Plugin-specific Notes
---------------------

- *Common Use Cases:* As more robust and more powerful replacement for the
  `~pyrseus.ctx.plugins.process` and `~pyrseus.ctx.plugins.cpprocess` plugins.

- *Concurrency:* Each worker runs in its own process. Unlike
  `~pyrseus.ctx.plugins.process` and `~pyrseus.ctx.plugins.cpprocess`, idle
  workers are automatically killed after a short timeout. This makes it more
  appropriate for interactive use and with workloads that have a low concurrent
  duty cycle.

- *Exceptions:* This plugin has standard exception-handling semantics: all
  task-related exceptions are captured in the task's future.

- *3rd Party Dependencies:* |cloudpickle|_ and |loky|_

- *Underlying Executors:* wrappers around various |loky|_ executors.

- *Default max_workers:* uses `~.loky.backend.context.cpu_count`, a smarter
  variant of our `~pyrseus.core.sys.get_num_available_cores`.

- *Pickling:* |cloudpickle|_

- *OnError handling:* Fully supports all `~pyrseus.ctx.api.OnError` modes.

  - To reduce dispatch latency, |loky|_  pre-queues many extra tasks, making
    them be uncancellable; so in `~pyrseus.ctx.api.OnError.CANCEL_FUTURES` mode,
    all of those pre-queued tasks are still run. If this results in too much
    latency, use `~pyrseus.ctx.api.OnError.KILL_WORKERS`.

  - `~pyrseus.ctx.api.OnError.CANCEL_FUTURES` is the default mode for reusable
    executors and `~pyrseus.ctx.api.OnError.KILL_WORKERS` is the default for
    non-reusable ones.

  - In reusable mode, `~pyrseus.ctx.api.OnError.KILL_WORKERS` will kill all
    workers in the pool, not just those working on tasks for one particular
    `~pyrseus.ctx.mgr.ExecutorCtx`. This is the only reason why
    `~pyrseus.ctx.api.OnError.CANCEL_FUTURES` is currently the default in
    reusable mode.

See :doc:`../plugins` for a summary of related plugins, and installation notes.
"""

from functools import cached_property
from typing import Optional, TypeVar

from pyrseus.core.sys import module_exists
from pyrseus.ctx.api import (
    ExecutorPluginEntryPoint,
    OnError,
    OnErrorLike,
    extract_keywords,
)
from pyrseus.ctx.simple import OnExitProxy

Ret = TypeVar("Ret")
"""
Represents the generic return type of a submitted callable.
"""


class EntryPoint(ExecutorPluginEntryPoint):
    supports_serial = False
    supports_concurrent = True

    @cached_property
    def is_available(self) -> bool:
        return module_exists("loky")

    @cached_property
    def allowed_keywords(self):
        from loky import ProcessPoolExecutor, get_reusable_executor

        return (
            extract_keywords(get_reusable_executor)
            | extract_keywords(ProcessPoolExecutor)
            | extract_keywords(self.create)
        )

    def create(
        self,
        max_workers: Optional[int] = None,
        *,
        reusable: bool = False,
        on_error: Optional[OnErrorLike] = None,
        **kwargs,
    ):
        """
        Creates a |loky|_ executor.

        As detailed below, this factory first asks |loky|_ for an appropriate
        executor instance. We then wrap that instance in `.OnExitProxy` to imbue
        it with the ``__exit__`` time behavior that other Pyrseus plugins
        provide.

        - If ``reusable`` is false, the executor is obtained via
          `.loky.ProcessPoolExecutor`.

          - If supplied, ``reuse`` and ``kill_workers`` keyword arguments are
            silently ignored (as is standard practice for unused keyword
            arguments with `~pyrseus.ctx.mgr.ExecutorCtx`).

          - If ``on_error`` is ``None``, it's treated as if it was
            `.OnError.CANCEL_FUTURES`. Regardless, ``on_error`` is consumed by
            this factory method.

          - All other arguments are passed to the underlying executor's
            constructor, as-is.

          - An `.OnExitProxy` will be created with ``shutdown=False``.

        - If ``reusable`` is true, the executor is obtained by
          `.loky.get_reusable_executor`.

          - If ``on_error`` is ``None``, it's treated as if it was
            `.OnError.KILL_WORKERS`. Regardless, ``on_error`` is consumed by
            this factory method.

          - All other arguments are passed to the underlying executor's
            constructor, as-is.

          - An `.OnExitProxy` will be created with ``shutdown=True``.

        :param max_workers: the maximum number of workers to use in the pool.
            This is passed directly to |loky|_ since it has a smart algorithm
            for avoiding oversubscription of CPU resources.

        :param reusable: decides whether to share a worker pool with other
            contexts or not. Reusable pools can have better startup time, but by
            default they can block for arbitrary amounts of time if there was an
            uncaught exception within the context.

        :param on_error: controls what to do with the submitted futures and the
            worker pool if there was an uncaught exception within the context.
            See `.OnExitProxy` and `.OnError` for details.
        """
        from loky import ProcessPoolExecutor, get_reusable_executor

        # Don't do anything special here when max_workers=None. loky's defaults
        # are even better than what we do in process and cpprocess, e.g. it also
        # tries to respect cgroup limits.

        if reusable:
            # CANCEL_FUTURES can be especially slow with loky reusable executors
            # because it pre-loads a large work queue with tasks. But it's a
            # safer approach in reusable mode. We choose to be safe here since
            # reusable mode is opt-in.
            if on_error is None:
                on_error = OnError.CANCEL_FUTURES
            else:
                on_error = OnError.get(on_error)
            # In reusable mode, we don't own the whole worker pool, so we setup
            # a proxy that gives the executor context manager semantics with
            # efforts taken to avoid interfering with futures that were not
            # submitted by this proxy instance.
            shutdown = False
            exe = get_reusable_executor(max_workers, **kwargs)
        else:
            # In non-reusable mode, we own the whole worker pool, so it's okay
            # to use shutdown=True. Alternatively, we could replicate the logic
            # in SimpleEntryPoint.create, but we choose to use OnExitProxy here
            # too for consistency.
            #
            # But first, we need to remove any reusable-only arguments. There
            # are only two right now, so we do it by hand.
            kwargs.pop("reuse", None)
            kwargs.pop("kill_workers", None)
            # It's safe enough to be aggressive since we own the whole pool.
            if on_error is None:
                on_error = OnError.KILL_WORKERS
            else:
                on_error = OnError.get(on_error)
            # This mimics what SimpleEntryPoint.create does, but specialized for
            # loky. Loky's shutdown method doesn't natively support cancel mode,
            # so we do still have to emulate that case.
            shutdown = on_error is not OnError.CANCEL_FUTURES
            ctx = ProcessPoolExecutor(max_workers, **kwargs)
            exe = ctx.__enter__()  # just to be pedantic

        proxy = OnExitProxy(exe, on_error=on_error, shutdown=shutdown)
        return proxy, None


ENTRY_POINT = EntryPoint()



# ---
# jupyter:
#   jupytext:
#     formats: ipynb,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.16.1
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown] editable=true slideshow={"slide_type": ""} tags=["pyflyby-cell"]
# # ExecutorCtx vs. Classes
#
# In this notebook, we explore how `ExecutorCtx` can make it easier to switch
# between different executor classes, relative to directly using those classes.

# %% [markdown]
# ## Setup
#
# Before running this notebook, make sure everything it depends on is installed:
#
#     # Modify the docs/requirements.txt path if you're running this
#     # command from anything except the repository's root directory.
#     python -m pip install -r docs/requirements.txt
#
# These are some imports that we'll use throughout this notebook:

# %% editable=true slideshow={"slide_type": ""} tags=["pyflyby-cell"]
import sys

sys.path.append("../../../src")  # assume we're running it from a Pyrseus source clone
import os
from socket import gethostname
from threading import get_ident

# %% [markdown]
# Here's a helper function that we'll use in each of the examples
# that features a valid executor.


# %%
def infer_exe_type(exe):
    """
    Runs a few tasks to infer what major type of executor is being used.
    """
    # We use multiple tasks in this notebook to avoid needing to
    # discuss pickle vs. cloudpickle issues.
    if exe.submit(gethostname).result() != gethostname():
        return "Using a multi-host executor."
    elif exe.submit(os.getpid).result() != os.getpid():
        return "Using a multi-process executor."
    elif exe.submit(get_ident).result() != get_ident():
        return "Using a multi-threaded executor."
    else:
        return "Using a serial executor."


# %% [markdown]
# ## Using Explicit Classes
#
# Suppose we have some initial concurrent code that uses `ProcessPoolExecutor`
# with a few constructor arguments.

# %%
# Reasonable initial code.
from concurrent.futures import ProcessPoolExecutor  # NOQA
from multiprocessing import get_context  # NOQA

with ProcessPoolExecutor(2, mp_context=get_context("spawn")) as exe:
    # All of the with-statement bodies are trivial in this notebook so
    # that we can concentrate on the different ways of constructing
    # the executors.
    print(infer_exe_type(exe))

# %% [markdown]
# Now suppose that we want to run that code serially. Then we need to change
# all of the setup lines, without making any mistakes:
# - change the imports,
# - change the executor class used in the with-statement,
# - remove the `max_workers` argument, and
# - remove the `mp_context` argument.

# %%
# Switching executors required a lot of changes.
from pyrseus.executors.inline import InlineExecutor  # NOQA

with InlineExecutor() as exe:
    print(infer_exe_type(exe))

# %% [markdown]
# If debugging, we might then start commenting out lines to switch back and
# forth, making it even messier e.g.:

# %%
# Switching back and forth while debugging can be cumbersome.
from concurrent.futures import ProcessPoolExecutor  # NOQA
from multiprocessing import get_context  # NOQA

from pyrseus.executors.inline import InlineExecutor  # NOQA

# with ProcessPoolExecutor(2, mp_context=get_context("spawn")) as exe:
with InlineExecutor() as exe:
    print(infer_exe_type(exe))

# %% [markdown]
# Or we might take the extra effort to do it one of the "right" ways, at the
# cost of yet more code clutter:

# %%
# Switching back and forth a "right" way clutters the code.
mode = "inline"  # manually edited while debugging to switch modes
from concurrent.futures import ProcessPoolExecutor  # NOQA
from contextlib import ExitStack  # NOQA
from multiprocessing import get_context  # NOQA

from pyrseus.executors.inline import InlineExecutor  # NOQA

with ExitStack() as es:
    # Complicated!
    if mode == "inline":
        exe = es.enter_context(InlineExecutor())
    elif mode == "process":
        exe = es.enter_context(ProcessPoolExecutor(2, mp_context=get_context("spawn")))
    else:
        raise ValueError(mode)
    # As before...
    print(infer_exe_type(exe))

# %% [markdown]
# ## Using ExecutorCtx with Explicit Plugins
#
# If we use `ExecutorCtx` instead, then we can keep the simplicity of the
# original code while also making it trivial to switch between different
# executor classes.

# %%
# ExecutorCtx variant of the initial code. This looks very
# similar to the "Explicit Classes" variant.
from multiprocessing import get_context  # NOQA

from pyrseus.ctx.mgr import ExecutorCtx  # NOQA

with ExecutorCtx("process", 2, mp_context=get_context("spawn")) as exe:
    print(infer_exe_type(exe))

# %% [markdown]
# Switching to `InlineExecutor` is just a matter of changing the first argument
# to `ExecutorCtx`. The context manager knows that `InlineExecutor` doesn't
# care about the positional `max_workers` argument or the keyword `mp_context`
# argument, so it silently hides them from `InlineExecutor`.

# %%
# All we had to change was the plugin name. Nothing else.
from multiprocessing import get_context  # NOQA

from pyrseus.ctx.mgr import ExecutorCtx  # NOQA

with ExecutorCtx("inline", 2, mp_context=get_context("spawn")) as exe:
    print(infer_exe_type(exe))

# %% [markdown]
# Alternatively, it would still be fine to remove those extra arguments if we
# don't anticipate going back and using `ProcessPoolExecutor` anymore.

# %%
# It's also okay to remove the extra arguments if we won't be using the
# "process" plugin anymore.
with ExecutorCtx("inline") as exe:
    print(infer_exe_type(exe))

# %% [markdown]
# `ExecutorCtx` is nice and will complain about keyword arguments that no
# available plugin accepts. Furthermore, it tells us what the available
# plugins and allowed keywords are. That way we can see if the problem
# is a missing plugin, a mistyped keyword, etc.

# %%
# But ExecutorCtx does help identify keywords that are not accepted by any
# plugin.
try:
    ExecutorCtx(
        "inline",
        # max_workers=2 is silently ignored because it could be relevant for a
        # concurrent plugin.
        2,
        # mp_context is silently ignored because some plugins like "process" can
        # use it.
        mp_context=get_context("spawn"),
        # But ExecutorCtx is nice and complains about this keyword that is
        # always ignored.
        not_a_keyword=42,
    )
except Exception as ex:
    print("CAUGHT EXCEPTION:", ex)

# %% [markdown]
# ## Using ExecutorCtx with Implicit Plugins
#
# Another way to use `ExecutorCtx` is to rely on its system for managing a
# default serial and a default concurrent plugin. If we just care about
# switching between those two defaults, then we just need to *exclude* the
# plugin name argument and change the `max_workers` argument.
#
# Here's the analogous initial code snippet.

# %%
# In this snippet, we just exclude "process".
from multiprocessing import get_context  # NOQA

from pyrseus.ctx.mgr import ExecutorCtx  # NOQA

with ExecutorCtx(2, mp_context=get_context("spawn")) as exe:
    print(infer_exe_type(exe))

# %% [markdown]
# Then by changing `2` to `0`, we'll implicitly use `InlineExecutor`, assuming
# it's the default.

# %%
# Now we're using InlineExecutor, just by changing 2 to 0.
from multiprocessing import get_context  # NOQA

from pyrseus.ctx.mgr import ExecutorCtx  # NOQA

with ExecutorCtx(0, mp_context=get_context("spawn")) as exe:
    print(infer_exe_type(exe))

# %% [markdown]
# See the "Plugins for ExecutorCtx" documentation page for the various ways
# of changing the default plugins.



"""
`~pyrseus.ctx.mgr.ExecutorCtx` plugin that provides a factory for the built-in
`~concurrent.futures.ThreadPoolExecutor` that respects the CPU affinity mask
when choosing the default thread count.

>>> import os
>>> from pyrseus.ctx.registry import skip_if_unavailable
>>> from pyrseus.ctx.mgr import ExecutorCtx

>>> # "thread" doesn't pickle, so lambdas are fine.
>>> skip_if_unavailable("thread")
>>> with ExecutorCtx("thread") as exe:
...     assert exe.__class__.__name__ == 'ThreadPoolExecutor'
...     assert exe.submit(lambda: os.getpid()).result() == os.getpid()

Plugin-specific Notes
---------------------

- *Common Use Cases:* as replacement for directly using
  `concurrent.futures.ThreadPoolExecutor` for users that enjoy using
  `~pyrseus.ctx.mgr.ExecutorCtx` and/or want to have a more conservative default
  ``max_workers``.

- *Concurrency:* Each worker owns its own thread within the main process.

- *Exceptions:* This plugin has standard exception-handling semantics: all
  task-related exceptions are captured in the task's future.

- *3rd Party Dependencies:* This plugin has no 3rd party dependencies.

- *Underlying Executor:* `concurrent.futures.ProcessPoolExecutor`

- *Default max_workers:* Uses `~pyrseus.core.sys.get_num_available_cores`
  instead of `multiprocessing.cpu_count`, respecting the CPU affinity mask when
  possible.

- *Pickling:* Not applicable. For troubleshooting pickling issues, consider
  using one of the serial executors instead, e.g.
  `~pyrseus.ctx.plugins.cpinline`.

- *OnError handling:* Fully supports `~pyrseus.ctx.api.OnError.WAIT` and
  `~pyrseus.ctx.api.OnError.CANCEL_FUTURES` modes.
  `~pyrseus.ctx.api.OnError.KILL_WORKERS` mode is automatically downgraded to
  `~pyrseus.ctx.api.OnError.CANCEL_FUTURES`.

See :doc:`../plugins` for a summary of related plugins, and installation notes.
"""

from pyrseus.ctx.simple import SimpleEntryPoint, shutdown_cancel_on_error


class EntryPoint(SimpleEntryPoint):
    supports_serial = False
    supports_concurrent = True
    executor_module_name = "concurrent.futures"
    executor_class_name = "ThreadPoolExecutor"

    # When there are no exceptions flowing through __exit__, the built-in
    # behavior is correct.
    wrap_for_wait_on_error = lambda self, ctx: (ctx, None)  # NOQA
    # When the user wants upgraded __exit__ semantics to cancel all pending
    # futures on exit, do so.
    wrap_for_cancel_on_error = lambda self, ctx: (ctx, shutdown_cancel_on_error)  # NOQA
    # Silently downgrade kill to cancel since there is no safe and generic way
    # to kill a thread.
    wrap_for_kill_on_error = lambda self, ctx: (ctx, shutdown_cancel_on_error)  # NOQA


ENTRY_POINT = EntryPoint()



"""
Tests the built-in `multiprocessing.Pool`'s robustness to crashes. This was
adapted from one of Loky's unit test files: `test_reusable_executor
<https://github.com/joblib/loky/blob/master/tests/test_reusable_executor.py>_.
That file is subject to the following license::

    BSD 3-Clause License

    Copyright (c) 2017, Olivier Grisel & Thomas Moreau.
    All rights reserved.

    Redistribution and use in source and binary forms, with or without
    modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice, this
    list of conditions and the following disclaimer.

    * Redistributions in binary form must reproduce the above copyright notice,
    this list of conditions and the following disclaimer in the documentation
    and/or other materials provided with the distribution.

    * Neither the name of the copyright holder nor the names of its
    contributors may be used to endorse or promote products derived from
    this software without specific prior written permission.

    THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
    AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
    IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
    DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
    FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
    DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
    SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
    CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
    OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
    OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import os.path
import platform
import subprocess
import sys
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple, Union

import pytest

from pyrseus.ctx.registry import skip_if_unavailable

# These tests intentionally trigger deadlocks in Python's multiprocessing
# library, so they're slow to run. We skip them all in CI runs.
#
# These tests rely on SIGALRM. Unfortunately, that doesn't exist on Windows.
if platform.system() == "Windows":
    pytestmark = [
        pytest.mark.skip("SIGALRM does not exist on Windows"),
        pytest.mark.slow,
    ]
else:
    pytestmark = pytest.mark.slow


@dataclass
class Case:
    func_name: str
    arg_name: str
    last_step: str
    errors: Optional[Union[str, Tuple[str]]] = None
    loky_errors: Optional[Union[str, Tuple[str]]] = None
    timeout_secs: int = 2

    @property
    def eff_errors(self):
        if self.errors is None:
            return ()
        elif isinstance(self.errors, str):
            return (self.errors,)
        elif isinstance(self.errors, tuple):
            return self.errors
        else:
            return TypeError(type(self.errors))


KNOWN_ERROR_STRINGS = (
    "BrokenProcessPool",
    "PicklingError",
    # "RemoteTraceback",
    "Segmentation fault",
    "SystemExit",
    "UnpicklingError",
)

CASES: List[Union[Case, Dict[Optional[str], Case]]] = [
    # fmt: off

    # verifies that the timeouts work
    Case("sleep", "zero", "EXITING-SCRIPT", timeout_secs=1),
    Case("sleep", "ten",  "TIMED-OUT",      timeout_secs=1),

    # Check problem occurring while pickling a task in
    Case("id", "error_at_pickle", "CAUGHT", ("RemoteTraceback", "PicklingError")),
    {
        'loky': Case("id", "exit_at_pickle", "CAUGHT", ('RemoteTraceback', 'SystemExit', 'PicklingError')),
        None:   Case("id", "exit_at_pickle", "TIMED-OUT"),
    },
    # The following two are disabled for now.
    #  - With "process", the test deadlocks, and it leaves 1 orphaned process
    #    even after manually killing the test.
    #  - With "loky", the same problems exist, but 3 orphans are left over.
    #
    # - Case("id", "c_exit_at_pickle", "TIMED-OUT"),
    # - Case("id", "crash_at_pickle",  "TIMED-OUT"),

    # Check problem occurring while unpickling a task on workers
    Case("id", "error_at_unpickle", "CAUGHT", ("UnpicklingError", "BrokenProcessPool")),
    {
        'loky': Case("id", "exit_at_unpickle", "CAUGHT", ("RemoteTraceback", "SystemExit", "BrokenProcessPool")),
        None:   Case("id", "exit_at_unpickle", "CAUGHT", "BrokenProcessPool"),
    },
    Case("id", "c_exit_at_unpickle", "CAUGHT", "BrokenProcessPool"),
    {
        'loky': Case("id", "crash_at_unpickle", "CAUGHT", "TerminatedWorkerError"),
        None:   Case("id", "crash_at_unpickle", "CAUGHT", "BrokenProcessPool"),
    },

    # Check problem occurring during function execution on workers
    Case("raise",  "runtime_error", "CAUGHT", ("RemoteTraceback", "RuntimeError")),
    Case("exit",   "na",            "CAUGHT", "SystemExit"),
    Case("c_exit", "na",            "CAUGHT", "BrokenProcessPool"),
    {
        'loky': Case("crash", "na", "CAUGHT", "TerminatedWorkerError"),
        None:   Case("crash", "na", "CAUGHT", "BrokenProcessPool"),
    },

    # Check problem occurring while pickling a task result on workers
    Case("invoke", "ctor_error_at_pickle",  "CAUGHT", ("RemoteTraceback", "PicklingError")),
    Case("invoke", "ctor_exit_at_pickle",   "CAUGHT", "SystemExit"),
    Case("invoke", "ctor_c_exit_at_pickle", "CAUGHT", "BrokenProcessPool"),
    {
        'loky': Case("invoke", "ctor_crash_at_pickle", "CAUGHT", "TerminatedWorkerError"),
        None:   Case("invoke", "ctor_crash_at_pickle", "CAUGHT", "BrokenProcessPool"),
    },

    # Check problem occurring while unpickling a task in the result handler
    Case("invoke", "ctor_error_at_unpickle", "CAUGHT", ("RemoteTraceback", "UnpicklingError", "BrokenProcessPool")),
    Case("invoke", "ctor_exit_at_unpickle",  "CAUGHT", ("RemoteTraceback", "SystemExit", "BrokenProcessPool")),
    # fmt: on
]


@pytest.mark.parametrize("case", list(range(len(CASES))))
@pytest.mark.parametrize("plugin", ("process", "loky"))
def test_crash_scenarios(plugin: str, case: int):
    skip_if_unavailable(plugin)

    case = CASES[case]
    if isinstance(case, dict):
        case = case.get(plugin, case[None])

    # Run the simple executor test.
    cmd = [
        sys.executable,
        "_crash_script.py",
        str(case.timeout_secs),
        plugin,
        case.func_name,
        case.arg_name,
    ]
    res = subprocess.run(
        cmd,
        cwd=os.path.dirname(__file__),
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT,
        timeout=case.timeout_secs + 5,  # an extra layer of timeout protection
        text=True,
    )

    # Quick checks and early exit if we expected it to succeed.
    if case.last_step == "EXITING-SCRIPT":
        assert res.returncode == 0
        assert res.stdout.rstrip().endswith("STEP: EXITING-SCRIPT")
        for s in KNOWN_ERROR_STRINGS:
            assert s not in res.stdout, s
        return

    # More involved checks for the normal case of expecting some kind of
    # failure.
    msgs = []

    last_step = None
    lines = res.stdout.splitlines()
    for line in lines:
        if line.startswith("STEP: "):
            last_step = line.split()[1]
    if last_step != case.last_step:
        msgs.append(f"Last step: expected={case.last_step} actual={last_step}")

    errors = case.eff_errors
    for error in errors:
        if error not in res.stdout:
            msgs.append(
                f"Expected {error} in the test script's output, "
                f"but it was not there."
            )

    for error in sorted(set(KNOWN_ERROR_STRINGS) - set(errors)):
        if error in res.stdout:
            msgs.append(f"Unexpected error {error} in the test script's output.")

    if msgs:
        msgs.insert(0, "Unexpected result")
        msgs.append("Command:")
        msgs.append(" ".join(cmd))
        msgs.append("Script output:")
        msgs.append(res.stdout)
        pytest.fail("\n".join(msgs))



"""
Tests various pickling and kwarg features of both `.ProcessPoolExecutor` and
`.CpProcessPoolExecutor`. This test case requires pytest or a compatible runner
due to the test parametrization we use.
"""

import inspect
import multiprocessing
import os
from collections import defaultdict
from concurrent.futures import ProcessPoolExecutor
from contextlib import ExitStack
from pickle import PickleError
from unittest import SkipTest

import pytest
from pyrseus.core.sys import is_mp_start_method_supported
from pyrseus.executors.cpprocess import CpProcessPoolExecutor

POOL_PARAMS = inspect.signature(ProcessPoolExecutor).parameters
"""
Parameters of the underlying ProcessPoolExecutor, for detecting features that
only exist in some versions of Python and/or on some platforms.
"""

PID = os.getpid()


def get_worker_id():
    return os.getpid()


def get_global_pid():
    return PID


def init_global_pid(pid):
    global PID
    PID = pid


def get_lambda():
    return lambda: 123


def _get_mp_context_or_skip(start_method):
    if start_method is None:
        mp_context = None
    elif "mp_context" not in POOL_PARAMS:
        raise SkipTest("mp_context is not supported in your interpreter")
    elif not is_mp_start_method_supported(start_method):
        raise SkipTest(f"{start_method=} is not supported in your interpreter")
    else:
        mp_context = multiprocessing.get_context(start_method)
    return mp_context


@pytest.mark.parametrize("start_method", (None, "fork", "forkserver", "spawn"))
def test_start_methods(start_method):
    # Check that our initial state is reasonable.
    fake_worker_id = -123
    assert get_worker_id() != fake_worker_id
    assert PID != fake_worker_id
    assert PID == os.getpid()

    # Create the mp_context.
    mp_context = _get_mp_context_or_skip(start_method)

    # Verify that the initializer ran for the worker process.
    with CpProcessPoolExecutor(
        1,
        mp_context=mp_context,
        initializer=init_global_pid,
        initargs=(fake_worker_id,),
    ) as exe:
        assert exe.submit(get_global_pid).result() == fake_worker_id

    # Make sure it didn't poison our process.
    assert PID != fake_worker_id

    # Make sure it didn't poison future workers.
    with CpProcessPoolExecutor(1, mp_context=mp_context) as exe:
        assert exe.submit(get_global_pid).result() != fake_worker_id

    # Ensure we didn't pollute and globals in this process.
    assert PID == os.getpid()


def test_max_tasks_per_child():
    if "max_tasks_per_child" not in POOL_PARAMS:
        raise SkipTest("max_tasks_per_child is not supported in your interpreter")
    # NOTE: until https://github.com/python/cpython/issues/115831 is resolved,
    # only use NUM_TASKS_PER_SET=1. Larger values will hang or crash the
    # process, due to bugs in the multiprocessing library.
    NUM_SETS = 5
    NUM_TASKS_PER_SET = 1
    with CpProcessPoolExecutor(max_tasks_per_child=NUM_TASKS_PER_SET) as exe:
        futs = [exe.submit(get_worker_id) for _ in range(NUM_SETS * NUM_TASKS_PER_SET)]
        counts = defaultdict(int)
        for fut in futs:
            counts[fut.result()] += 1
        assert len(counts) == NUM_SETS
        for pid, count in counts.items():
            assert count == NUM_TASKS_PER_SET, (pid, count)


@pytest.mark.parametrize("start_method", (None, "fork", "forkserver", "spawn"))
@pytest.mark.parametrize(
    "cls,should_succeed,func",
    (
        (ProcessPoolExecutor, True, get_worker_id),
        (ProcessPoolExecutor, False, lambda: get_worker_id),
        (ProcessPoolExecutor, False, get_lambda),
        (CpProcessPoolExecutor, True, get_worker_id),
        (CpProcessPoolExecutor, True, lambda: get_worker_id),
        (CpProcessPoolExecutor, True, get_lambda),
    ),
)
def test_pickling(cls, should_succeed, func, start_method):
    """
    Verifies that

     - ProcessPoolExecutor can only pickle normally-picklable things (mostly to
       show that the CpProcessPoolExecutor tests are actually testing something
       real

     - CpProcessPoolExecutor can handle bi-directional pickling that requires
       cloudpickle.
    """
    mp_context = _get_mp_context_or_skip(start_method)
    with ExitStack() as es:
        exe = es.enter_context(cls(1, mp_context=mp_context))
        if not should_succeed:
            # Trap errors from the library, not assertions from this test.
            es.enter_context(pytest.raises((AttributeError, RuntimeError, PickleError)))
        fut = exe.submit(func)
        ret = fut.result()
        if func is get_lambda:
            assert ret() == 123
        else:
            assert ret != get_worker_id()



"""
Provides a simple serial executor that does *not* capture exceptions in
submitted futures. This is primarily useful for troubleshooting when one wants
to enter a debugger as early and as easily as possible.
"""

from concurrent.futures import Executor, Future
from typing import Callable, TypeVar

__all__ = ["NoCatchExecutor"]

# Represents the generic return type of a submitted callable.
Ret = TypeVar("Ret")


class NoCatchExecutor(Executor):
    def __init__(self):
        """
        Creates a simple serial `~concurrent.futures.Executor` that evaluates
        tasks immediately upon submission, and does *not* capture task
        exceptions in their futures.

        This is primarily useful for troubleshooting when one wants to enter a
        debugger as early and as easily as possible, at the cost of non-standard
        error handling.

        Consider the following function that raises an exception:

            >>> def raises():
            ...     raise RuntimeError("An exception was raised by our function.")

        With this class, the exception is propagated out immediately at
        `~NoCatchExecutor.submit` time.

            >>> with NoCatchExecutor() as exe:
            ...     exe.submit(raises)  # <--- NOTE: no fut.result() needed
            Traceback (most recent call last):
            ...
            RuntimeError: An exception was raised by our function.

        See :doc:`../plugins` for a list of related executors.
        """
        self._closing = False

    def submit(self, fcn: Callable[..., Ret], /, *args, **kwargs) -> Future[Ret]:
        """
        Immediately evaluates ``fcn(*args, **kwargs)`` and embeds the result in
        a `~concurrent.futures.Future`. Unlike with standard executors, this
        method does *not* capture exceptions. They are propagated out
        immediately.
        """
        # This implementation is the same as InlineExecutor.submit, except this
        # one does not have a try-except-else construct. This breaks the normal
        # `~concurrent.futures.Executor` protocol, but it can make debugging
        # easier if users want exceptions to flow through to the default handler
        # instead of being captured by the returned
        # `~concurrent.futures.Future`.

        # To be consistent with `~concurrent.futures.ProcessPoolExecutor`,
        # disallow new submissions once shutdown has started.
        if self._closing:
            raise RuntimeError(
                "Submissions are not allowed to an executor that is shutting down."
            )

        # Eagerly execute the function *without* a try-except guard. Users who
        # want the standard guard should use `.InlineExecutor` instead of
        # `.NoCatchExecutor`.
        result = fcn(*args, **kwargs)

        # Handle the result. In this class we never call
        # `fut.set_exception(exc)` because we propagate exceptions back to the
        # user instead of capturing them.
        fut: Future[Ret] = Future()
        fut.set_result(result)

        return fut

    def shutdown(self, *args, **kwargs):
        """
        As a minor improvement on the base class' method, this override
        disallows submissions after a shutdown has been started. This can assist
        with finding bugs in user code.
        """
        self._closing = True
        super().shutdown(*args, **kwargs)



"""
Tests some edge cases of `.extract_keywords`.
"""

import pytest
from pyrseus.ctx.api import extract_keywords

#
# Call signature types that we support.
#


def test_max_workers_positional_only_okay():
    def factory(max_workers, /, foo=1, bar=2):
        pass

    assert extract_keywords(factory) == {"foo", "bar"}


def test_max_workers_dual_use_okay():
    def factory(max_workers, *, foo=1, bar=2):
        pass

    assert extract_keywords(factory) == {"foo", "bar"}


def test_unrestricted_okay():
    def factory(max_workers, foo=1, bar=2):
        pass

    assert extract_keywords(factory) == {"foo", "bar"}


def test_max_workers_with_default_okay():
    def factory(max_workers=0, foo=1, bar=2):
        pass

    assert extract_keywords(factory) == {"foo", "bar"}


def test_extra_dual_use_with_default_okay():
    def factory(max_workers, /, foo=1, *, bar=2):
        pass

    assert extract_keywords(factory) == {"foo", "bar"}


def test_self_okay():
    def factory(self, max_workers, /, foo=1, *, bar=1):
        pass

    assert extract_keywords(factory) == {"foo", "bar"}


class BadBase:
    # Invalid call signature in this base class, but it's okay because one of
    # our subclasses blocks the recursion by not taking a **kwargs arg.
    def __init__(self, **kwargs):
        pass


class BlockingSub(BadBase):
    # This subclass' lack of a **kwargs arg stops BadBase from breaking the
    # test.
    def __init__(self, max_workers, a=1, b=2):
        pass


class KwSub(BlockingSub):
    # An intermediate subclass that takes a **kwargs arg.
    def __init__(self, max_workers, b=3, c=4, **kwargs):
        pass


class Concrete(KwSub):
    # A final subclass that takes a **kwargs arg.
    def __init__(self, max_workers, *, d=5, **kwargs):
        pass


def test_complex_inheritance():
    assert extract_keywords(Concrete) == {"a", "b", "c", "d"}

    with pytest.raises(TypeError, match="first .* max_workers"):
        extract_keywords(BadBase)


#
# Call signature types that we do *not* support.
#


def test_extra_dual_use_non_default_fails():
    def factory(max_workers, /, foo, *, bar=1):
        pass

    with pytest.raises(TypeError, match="Parameter 'foo' has no default."):
        assert extract_keywords(factory) == {"foo", "bar"}


def test_max_workers_kw_only_fails():
    def factory(*, max_workers=1, foo=1, bar=2):
        pass

    # For now, we require that max_workers can be supplied positionally.
    with pytest.raises(TypeError, match="must be positional .* max_workers"):
        extract_keywords(factory)


def test_extra_positional_fails():
    def factory(max_workers, foo, /, *, bar=1):
        pass

    # No extra positional-only arguments are allowed.
    with pytest.raises(TypeError, match="The only allowed positional-only"):
        extract_keywords(factory)



"""
`~pyrseus.ctx.mgr.ExecutorCtx` plugin that provides a factory for the built-in
`~concurrent.futures.ProcessPoolExecutor` that respects the CPU affinity mask
when choosing the default worker process count.

>>> import os
>>> from pyrseus.ctx.registry import skip_if_unavailable
>>> from pyrseus.ctx.mgr import ExecutorCtx

>>> # "process" works with normally-picklable tasks and return values
>>> skip_if_unavailable("process")
>>> with ExecutorCtx("process", 1) as exe:
...     assert exe.submit(os.getpid).result() != os.getpid()

>>> # But "process" doesn't use cloudpickle like "cpprocess" does,
>>> # so it can't handle fancy things like lambdas.
>>> skip_if_unavailable("process")
>>> with ExecutorCtx("process") as exe:
...     exe.submit(lambda: os.getpid()).result()
Traceback (most recent call last):
...
_pickle.PicklingError: Can't pickle ...

Plugin-specific Notes
---------------------

- *Common Use Cases:* As replacement for directly using
  `concurrent.futures.ProcessPoolExecutor` for users that enjoy using
  `~pyrseus.ctx.mgr.ExecutorCtx` and/or want to have a more conservative default
  ``max_workers``.

- *Concurrency:* Each worker runs in its own process.

- *Exceptions:* This plugin has standard exception-handling semantics: all
  task-related exceptions are captured in the task's future.

- *3rd Party Dependencies:* This plugin has no 3rd party dependencies.

- *Underlying Executor:* `concurrent.futures.ProcessPoolExecutor`

- *Default max_workers*:* Uses `~pyrseus.core.sys.get_num_available_cores`
  instead of `multiprocessing.cpu_count`, respecting the CPU affinity mask when
  possible. Currently is unaware of cgroups constraints.

- *Pickling:* `pickle`.

- *OnError handling:* Fully supports `~pyrseus.ctx.api.OnError.WAIT` and
  `~pyrseus.ctx.api.OnError.CANCEL_FUTURES` modes.
  `~pyrseus.ctx.api.OnError.KILL_WORKERS` mode is automatically downgraded to
  `~pyrseus.ctx.api.OnError.CANCEL_FUTURES`.

  - By default, `~concurrent.futures.ProcessPoolExecutor`, pre-queues one extra
    task, making it uncancellable; so in
    `~pyrseus.ctx.api.OnError.CANCEL_FUTURES` mode, that pre-queued task will
    still be run.

.. note::

   If you can afford to add two extra 3rd party dependencies, consider using the
   `~pyrseus.ctx.plugins.loky` plugin instead of this one. Here are a few of its
   advantages:

   - Like our `~pyrseus.ctx.plugins.cpprocess` plugin, it uses the more powerful
     |cloudpickle|_ for pickling instead of just `pickle`.

   - Its `~.loky.backend.context.cpu_count` function for determining the default
     ``max_workers`` is more sophisticated than ours, e.g. it respects not only
     the CPU affinity mask, but also cgroups constraints.

   - Even though the |loky|_ project has already led to significant improvements
     to the built-in `multiprocessing` library, its executors still handle a few
     worker crash situations better than the built-in
     `~concurrent.futures.ProcessPoolExecutor` that this plugin uses (as of
     Python 3.10).

   - It supports reusable worker pools that are especially convenient to use in
     interactive code.

   - It supports the `~pyrseus.ctx.api.OnError.KILL_WORKERS` mode for faster
     teardowns when there are uncaught exceptions in the main program.

See :doc:`../plugins` for a summary of related plugins, and installation notes.
"""


import sys
from concurrent.futures import ProcessPoolExecutor
from functools import cached_property
from typing import Optional

from pyrseus.core.sys import get_num_available_cores
from pyrseus.ctx.api import (
    ExecutorPluginEntryPoint,
    OnError,
    OnErrorLike,
    extract_keywords,
)
from pyrseus.ctx.simple import OnExitProxy, shutdown_cancel_on_error


class EntryPoint(ExecutorPluginEntryPoint):
    supports_serial = False
    supports_concurrent = True
    is_available = True

    ExecutorCls = ProcessPoolExecutor

    @cached_property
    def allowed_keywords(self):
        return extract_keywords(self.create) | extract_keywords(self.ExecutorCls)

    def create(
        self,
        max_workers: Optional[int] = None,
        *,
        on_error: OnErrorLike = OnError.KILL_WORKERS,
        **kwargs,
    ):
        # Preprocess args.
        if max_workers is None:
            max_workers = max(1, get_num_available_cores())
        on_error = OnError.get(on_error)

        # ProcessPoolExecutor doesn't have a public API for cleanly killing
        # workers, so we downgrade on_error to cancel mode.
        if on_error is OnError.KILL_WORKERS:
            on_error = OnError.CANCEL_FUTURES

        # Create, based on the desired on_error mode.
        if on_error is OnError.WAIT:
            ctx = self.ExecutorCls(max_workers, **kwargs)
            pre_exit = None
        elif on_error is OnError.CANCEL_FUTURES:
            # Unfortunately, under some conditions shutdown(cancel_futures=True)
            # causes deadlocks on CPython <3.12
            # (https://github.com/python/cpython/issues/116685). So we fallback
            # to emulating is behavior in our code.
            if sys.version_info < (3, 12, 2):
                orig_ctx = self.ExecutorCls(max_workers, **kwargs)
                exe = orig_ctx.__enter__()
                ctx = OnExitProxy(exe, on_error=on_error, shutdown=False)
                pre_exit = None
            else:
                ctx = self.ExecutorCls(max_workers, **kwargs)
                pre_exit = shutdown_cancel_on_error
        else:
            raise ValueError(on_error)

        return ctx, pre_exit


ENTRY_POINT = EntryPoint()



"""
`~pyrseus.ctx.mgr.ExecutorCtx` plugin that provides a wrapper around the 3rd
party `~mpi4py.futures.MPIPoolExecutor`, enhancing it to use |cloudpickle|_ for
pickling tasks and their return values, instead of `pickle`.

>>> import os
>>> from pyrseus.ctx.registry import skip_if_unavailable
>>> from pyrseus.ctx.mgr import ExecutorCtx

>>> # "cpmpi4py" can pickle fancy things like lambdas with cloudpickle
>>> skip_if_unavailable("cpmpi4py")
>>> with ExecutorCtx("cpmpi4py", 1) as exe:
...     assert exe.submit(lambda: os.getpid()).result() != os.getpid()

Note that unlike `multiprocessing`-based executors, the |mpi4py|_-based ones
only replicate the ``PYTHONPATH`` environment variable by default in workers,
not `sys.path`. If you have made local modifications to your `sys.path` list, be
sure to replicate them via the ``path`` argument that's passed to
`~.mpi4py.futures.MPIPoolExecutor`.

Plugin-specific Notes
---------------------

- *Common Use Cases:* For multi-process and multi-host workloads, especially
  for for organizations that already have MPI setup but want a simple map-only
  interface to it.

- *Concurrency:* determined by the user's MPI configuration.

- *Exceptions:* This plugin has standard exception-handling semantics: all
  task-related exceptions are captured in the task's future.

- *3rd Party Dependencies:* |cloudpickle|_, |mpi4py|_

- *Underlying Executor:* a thin wrapper around
  `.mpi4py.futures.MPIPoolExecutor` that makes it use |cloudpickle|_ instead of
  `pickle` for pickling tasks and their results.

- *Default max_workers:* determined by the user's MPI configuration. Often this
  will require giving an explicit limit instead of relying on a pre-configured
  default.

- *Pickling:* |cloudpickle|_

- *OnError handling:* Fully supports `~pyrseus.ctx.api.OnError.WAIT` and
  `~pyrseus.ctx.api.OnError.CANCEL_FUTURES` modes.
  `~pyrseus.ctx.api.OnError.KILL_WORKERS` mode is automatically downgraded to
  `~pyrseus.ctx.api.OnError.CANCEL_FUTURES`.

See :doc:`../plugins` for a summary of related plugins, and installation notes.
"""

import multiprocessing
from functools import cache, cached_property
from typing import Optional

from pyrseus.core.pickle import CustomPickledClosure
from pyrseus.core.sys import module_exists
from pyrseus.ctx.api import ExecutorPluginEntryPoint, OnError, OnErrorLike
from pyrseus.ctx.simple import OnExitProxy


@cache
def get_class():
    """
    Lazily creates the underlying executor class. This is done in a function
    because plugins must not import 3rd party dependencies at module load time.

    Note: we don't have to worry about the picklability of the executor class:
    executors are typically not picklable anyway.
    """
    # Deferring these imports is the only reason we lazily create the class in
    # this function.
    import cloudpickle
    from mpi4py.futures import MPIPoolExecutor

    class CpMPIPoolExecutor(MPIPoolExecutor):
        def submit(self, fcn, /, *args, **kwargs):
            closure = CustomPickledClosure(
                fcn,
                args,
                kwargs,
                dumps=cloudpickle.dumps,
                loads=cloudpickle.loads,
            )
            return super().submit(closure)

    return CpMPIPoolExecutor


def init_mp_authkey(authkey):
    multiprocessing.current_process().authkey = authkey


def get_mp_authkey():
    return multiprocessing.current_process().authkey


class EntryPoint(ExecutorPluginEntryPoint):
    supports_serial = False
    supports_concurrent = True

    @cached_property
    def is_available(self) -> bool:
        return module_exists("mpi4py.futures") and module_exists("cloudpickle")

    allowed_keywords = {
        # Those supported by mpi4py. Unfortunately, their executor constructor
        # takes **kwargs, so we can't use metaprogramming like we do for other
        # plugins like process and loky.
        "python_exe",
        "python_args",
        "mpi_info",
        "globals",
        "main",
        "path",
        "wdir",
        "env",
    }

    def create(
        self,
        max_workers: Optional[int] = None,
        on_error: OnErrorLike = OnError.KILL_WORKERS,
        **kwargs,
    ):
        on_error = OnError.get(on_error)
        if on_error is OnError.KILL_WORKERS:
            # For now, demote KILL_WORKERS to CANCEL_FUTURES. MPIPoolExecutor
            # doesn't directly support a kill_workers mode. With the help of an
            # MPI expert, we'll be happy to consider adding a kill feature once
            # we know the right way to do it with MPI.
            on_error = OnError.CANCEL_FUTURES

        cls = get_class()
        ctx = cls(max_workers, **kwargs)

        if on_error is OnError.WAIT:
            return ctx, None
        elif on_error is OnError.CANCEL_FUTURES:
            exe = ctx.__enter__()  # just to be pedantic
            return OnExitProxy(exe, on_error=on_error, shutdown=True), None
        else:
            raise NotImplementedError(on_error)


ENTRY_POINT = EntryPoint()



#!/bin/env python3
"""
Helper script for testing the reliability of various multiprocessing backends,
especially ProcessPoolExecutor's multiprocessing.Pool backend. To run this
directly, first run the tests using ``tox``, then you can get help as follows::

    % ./.tox/py/bin/python tests/_crash_script.py --help

See unit tests that run this script for what the expected output is.

.. note:

    This script was adapted from one of Loky's unit test files:
    `test_reusable_executor
    <https://github.com/joblib/loky/blob/master/tests/test_reusable_executor.py>_.
    That file is subject to the following license::

        BSD 3-Clause License

        Copyright (c) 2017, Olivier Grisel & Thomas Moreau. All rights reserved.

        Redistribution and use in source and binary forms, with or without
        modification, are permitted provided that the following conditions are
        met:

        * Redistributions of source code must retain the above copyright notice,
          this list of conditions and the following disclaimer.

        * Redistributions in binary form must reproduce the above copyright
          notice, this list of conditions and the following disclaimer in the
          documentation and/or other materials provided with the distribution.

        * Neither the name of the copyright holder nor the names of its
          contributors may be used to endorse or promote products derived from
          this software without specific prior written permission.

        THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS
        IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
        TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
        PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
        HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
        SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED
        TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR
        PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
        LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING
        NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
        SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
"""

import argparse
import ctypes
import faulthandler
import os
import platform
import signal
import sys
import time
from multiprocessing import current_process
from pickle import PicklingError, UnpicklingError

import psutil

from pyrseus.ctx.mgr import ExecutorCtx

# Compat windows
if platform.system() == "Windows":
    from signal import SIGTERM as SIGKILL

    libc = ctypes.cdll.msvcrt
else:
    from ctypes.util import find_library
    from signal import SIGKILL

    libc = ctypes.CDLL(find_library("c"))


def crash():
    """Induces a segfault"""
    faulthandler._sigsegv()


def c_exit(exitcode=0):
    """Induces a libc exit with exitcode 0"""
    libc.exit(exitcode)


def raise_error(etype=UnpicklingError, message=None):
    """Function that raises an Exception in process"""
    raise etype(message)


def invoke(cls):
    """Function that returns a instance of cls"""
    return cls()


class CrashAtPickle:
    """Bad object that triggers a segfault at pickling time."""

    def __reduce__(self):
        crash()


class CrashAtUnpickle:
    """Bad object that triggers a segfault at unpickling time."""

    def __reduce__(self):
        return crash, ()


class ExitAtPickle:
    """Bad object that triggers a segfault at pickling time."""

    def __reduce__(self):
        exit()


class ExitAtUnpickle:
    """Bad object that triggers a process exit at unpickling time."""

    def __reduce__(self):
        return exit, ()


class CExitAtPickle:
    """Bad object that triggers a segfault at pickling time."""

    def __reduce__(self):
        c_exit()


class CExitAtUnpickle:
    """Bad object that triggers a process exit at unpickling time."""

    def __reduce__(self):
        return c_exit, ()


class ErrorAtPickle:
    """Bad object that raises an error at pickling time."""

    def __init__(self, fail=True):
        self.fail = fail

    def __reduce__(self):
        if self.fail:
            raise PicklingError("Error in pickle")
        else:
            return id, (42,)


class ErrorAtUnpickle:
    """Bad object that triggers a process exit at unpickling time."""

    def __init__(self, etype=UnpicklingError, message="the error message"):
        self.etype = etype
        self.message = message

    def __reduce__(self):
        return raise_error, (self.etype, self.message)


class CrashAtGCInWorker:
    """Bad object that triggers a segfault at call item GC time"""

    def __del__(self):
        if current_process().name != "MainProcess":
            crash()


class CExitAtGCInWorker:
    """Exit worker at call item GC time"""

    def __del__(self):
        if current_process().name != "MainProcess":
            c_exit()


def print_step(msg, *, tag="STEP"):
    print(f"{tag}: {msg}")
    sys.stdout.flush()


def timeout_handler(signum, frame):
    if signum == signal.SIGALRM:
        print_step("RECEIVED-SIGALRM")
        parent = psutil.Process(os.getpid())
        children = parent.children(recursive=True)
        for p in children:
            try:
                p.send_signal(SIGKILL)
            except psutil.NoSuchProcess:
                pass
        psutil.wait_procs(children)
        print_step("TIMED-OUT")
        c_exit()


FUNCS = {
    "c_exit": c_exit,
    "crash": crash,
    "exit": exit,
    "id": id,
    "invoke": invoke,
    "raise": raise_error,
    "sleep": time.sleep,
}

ARG_LISTS = {
    # Created on the main process. For use with "id".
    "c_exit_at_pickle": (CExitAtPickle(),),
    "c_exit_at_unpickle": (CExitAtUnpickle(),),
    "crash_at_pickle": (CrashAtPickle(),),
    "crash_at_unpickle": (CrashAtUnpickle(),),
    "error_at_pickle": (ErrorAtPickle(),),
    "error_at_unpickle": (ErrorAtUnpickle(),),
    "exit_at_pickle": (ExitAtPickle(),),
    "exit_at_unpickle": (ExitAtUnpickle(),),
    # Invoked on workers. For use with "invoke".
    "ctor_c_exit_at_pickle": (CExitAtPickle,),
    "ctor_c_exit_at_unpickle": (CExitAtUnpickle,),
    "ctor_crash_at_pickle": (CrashAtPickle,),
    "ctor_crash_at_unpickle": (CrashAtUnpickle,),
    "ctor_error_at_pickle": (ErrorAtPickle,),
    "ctor_error_at_unpickle": (ErrorAtUnpickle,),
    "ctor_exit_at_pickle": (ExitAtPickle,),
    "ctor_exit_at_unpickle": (ExitAtUnpickle,),
    # Others.
    "na": (),
    "runtime_error": (RuntimeError,),
    "ten": (10,),
    "zero": (0,),
}


def main(argv=None):
    parser = argparse.ArgumentParser()
    parser.add_argument("timeout_secs", type=int)
    parser.add_argument("plugin")
    parser.add_argument("func_name", choices=FUNCS)
    parser.add_argument("arg_name", choices=ARG_LISTS)
    parser.add_argument("--max-workers", type=int, default=1)
    args = parser.parse_args(argv)
    func = FUNCS[args.func_name]
    func_args = ARG_LISTS[args.arg_name]
    print_step("PARSED")
    signal.signal(signal.SIGALRM, timeout_handler)
    signal.alarm(args.timeout_secs)
    print_step("SET-TIMEOUT")
    ctx = ExecutorCtx(args.plugin, args.max_workers)
    print_step("CONSTRUCTED")
    with ctx as exe:
        print_step("ENTERED-CTX")
        try:
            print_step("ENTERED-TRY")
            fut = exe.submit(func, *func_args)
            print_step("SUBMITTED")
            res = fut.result()
        except Exception:
            print_step("CAUGHT")
            raise  # prints a traceback
        except SystemExit as ex:
            print_step(f"CAUGHT {type(ex).__name__}")
            raise  # no traceback printed for SystemExit
        except BaseException as ex:
            print_step(f"CAUGHT {type(ex).__name__}")
            raise  # presumably prints a traceback
        else:
            print_step(f"RESULT {res}")
    print_step("EXITED-CTX")


if __name__ == "__main__":
    main()
    print_step("EXITING-SCRIPT")



"""
Tests that overriding ``dumps`` and ``loads`` in the `pyrseus.core.pickle`
classes and functions works as expected.
"""

import pickle
from unittest import TestCase

from pyrseus.core.pickle import (
    CustomPickledClosure,
    OncePickledObject,
    call_with_round_trip_pickling,
    try_pickle_round_trip,
)

LOG = []


def logged_dumps(obj, protocol=-1):
    pickled = pickle.dumps(obj, protocol=protocol)
    LOG.append("pickled")
    return pickled


def logged_loads(obj):
    ret = pickle.loads(obj)
    LOG.append("unpickled")
    return ret


class LogResettingMixin:
    def setUp(self):
        LOG[:] = []

    def tearDown(self):
        LOG[:] = []


def add_10(x):
    return x + 10


class TestTroubleshootingFuncs(LogResettingMixin, TestCase):
    """
    Verifies that `try_pickle_round_trip` and `call_with_round_trip_pickling`
    honor their ``dumps`` and ``loads`` arguments.
    """

    def test_try_pickle_round_trip(self):
        assert LOG == []
        ret = try_pickle_round_trip(
            add_10,
            dumps=logged_dumps,
            loads=logged_loads,
        )
        assert ret is add_10
        assert LOG == ["pickled", "unpickled"]

    def test_call_with_round_trip_pickling(self):
        assert LOG == []
        closure_args = (add_10, (1,), {})
        ret = call_with_round_trip_pickling(
            *closure_args,
            dumps=logged_dumps,
            loads=logged_loads,
        )
        assert ret == 11
        assert LOG == ["pickled", "unpickled", "pickled", "unpickled"]


class TestOncePickledObject(LogResettingMixin, TestCase):
    """
    Verifies that `OncePickledObject` interacts with picklers in the expected
    way.
    """

    def test_trigger_timing(self):
        assert LOG == []

        opo = OncePickledObject(42, logged_dumps, logged_loads)
        assert LOG == []

        pickled = pickle.dumps(opo, -1)
        assert LOG == ["pickled"]

        reconstructed = pickle.loads(pickled)
        assert LOG == ["pickled", "unpickled"]
        assert isinstance(reconstructed, int)  # not OncePickledObject
        assert reconstructed == 42


class TestCustomPickledClosure(LogResettingMixin, TestCase):
    """
    Verifies that `CustomPickledClosure` interacts with picklers in the expected
    way.
    """

    def test_trigger_timing(self):
        assert LOG == []

        closure = CustomPickledClosure(
            add_10,
            (2,),
            {},
            logged_dumps,
            logged_loads,
        )
        assert LOG == []

        pickled_closure = pickle.dumps(closure, -1)
        assert LOG == ["pickled"]

        reconstructed_closure = pickle.loads(pickled_closure)
        assert LOG == ["pickled", "unpickled"]
        assert callable(reconstructed_closure)

        wrapped_ret = reconstructed_closure()
        assert LOG == ["pickled", "unpickled"]
        assert isinstance(wrapped_ret, OncePickledObject)  # not int

        pickled_ret = pickle.dumps(wrapped_ret, -1)
        assert LOG == ["pickled", "unpickled", "pickled"]

        unwrapped_ret = pickle.loads(pickled_ret)
        assert LOG == ["pickled", "unpickled", "pickled", "unpickled"]
        assert unwrapped_ret == 12
